# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""API routes for Image tasks."""

from typing import Annotated

from fastapi import APIRouter, Depends

from pixano_inference.pydantic import (
    CeleryTask,
    ImageMaskGenerationRequest,
    ImageMaskGenerationResponse,
    ImageZeroShotDetectionResponse,
)
from pixano_inference.pydantic.tasks.image.zero_shot_detection import ImageZeroShotDetectionRequest
from pixano_inference.routers.utils import delete_task, execute_task_request, get_task_result
from pixano_inference.settings import Settings, get_pixano_inference_settings
from pixano_inference.tasks import ImageTask


router = APIRouter(prefix="/image", tags=["Image"])


@router.post("/mask_generation/", response_model=CeleryTask)
async def mask_generation(
    request: ImageMaskGenerationRequest,
    settings: Annotated[Settings, Depends(get_pixano_inference_settings)],
) -> CeleryTask:
    """Generate mask from an image and optionnaly points and bboxes.

    Args:
        request: Request for image mask generation.
        settings: Settings of the app.

    Returns:
        Id and status of the task.
    """
    return await execute_task_request(request=request, task=ImageTask.MASK_GENERATION, settings=settings)


@router.get("/mask_generation/{task_id}", response_model=ImageMaskGenerationResponse | CeleryTask)
async def get_mask_generation(task_id: str) -> ImageMaskGenerationResponse | CeleryTask:
    """Get the result of a image mask generation task.

    Args:
        task_id: ID of the task to retrieve.

    Returns:
        Response for image mask generation.
    """
    return await get_task_result(task_id, response_type=ImageMaskGenerationResponse)


@router.delete("/mask_generation/{task_id}", response_model=None)
async def delete_mask_generation(task_id: str) -> None:
    """Delete a image mask generation task."""
    return await delete_task(task_id=task_id)


@router.post("/zero_shot_detection/", response_model=CeleryTask)
async def zero_shot_detection(
    request: ImageZeroShotDetectionRequest,
    settings: Annotated[Settings, Depends(get_pixano_inference_settings)],
) -> CeleryTask:
    """Perform zero shot detection on an image.

    Args:
        request: Request for image zero shot detection.
        settings: Settings of the app.

    Returns:
        Id and status of the task.
    """
    return await execute_task_request(request=request, task=ImageTask.ZERO_SHOT_DETECTION, settings=settings)


@router.get("/zero_shot_detection/{task_id}", response_model=ImageZeroShotDetectionResponse | CeleryTask)
async def get_zero_shot_detection(task_id: str) -> ImageZeroShotDetectionResponse | CeleryTask:
    """Get the result of a image zero shot detection task.

    Args:
        task_id: ID of the task to retrieve.

    Returns:
        Response for image zero shot detection.
    """
    return await get_task_result(task_id, response_type=ImageZeroShotDetectionResponse)


@router.delete("/zero_shot_detection/{task_id}", response_model=None)
async def delete_zero_shot_detection(task_id: str) -> None:
    """Delete a image zero shot detection task."""
    return await delete_task(task_id=task_id)



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Inference model for the SAM2 model."""

from __future__ import annotations

import gc
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import numpy as np
from PIL.Image import Image

from pixano_inference.pydantic.nd_array import NDArrayFloat
from pixano_inference.pydantic.tasks.image.mask_generation import ImageMaskGenerationOutput
from pixano_inference.pydantic.tasks.image.utils import CompressedRLE
from pixano_inference.pydantic.tasks.video.mask_generation import VideoMaskGenerationOutput
from pixano_inference.utils.package import assert_sam2_installed, is_sam2_installed, is_torch_installed

from .base import BaseInferenceModel


if is_torch_installed():
    import torch

if is_sam2_installed():
    from sam2.sam2_image_predictor import SAM2ImagePredictor
    from sam2.sam2_video_predictor import SAM2VideoPredictor
    from torch import Tensor

if TYPE_CHECKING:
    from torch import Tensor


class Sam2Model(BaseInferenceModel):
    """Inference model for the SAM2 model."""

    def __init__(
        self,
        name: str,
        provider: str,
        predictor: Any,
        torch_dtype: Literal["float32", "float16", "bfloat16"] = "bfloat16",
        config: dict[str, Any] = {},
    ):
        """Initialize the model.

        Args:
            name: Name of the model.
            provider: Provider of the model.
            predictor: The SAM2 image predictor.
            torch_dtype: The torch data type to use for inference.
            config: Configuration for the model.
        """
        assert_sam2_installed()

        super().__init__(name, provider)
        match torch_dtype:
            case "float32":
                self.torch_dtype = torch.float32
            case "float16":
                self.torch_dtype = torch.float16
            case "bfloat16":
                self.torch_dtype = torch.bfloat16
            case _:
                raise ValueError(f"Invalid torch_dtype: {torch_dtype}")
        self.predictor: SAM2ImagePredictor | SAM2VideoPredictor = predictor
        self.config = config

    def delete(self) -> None:
        """Delete the model."""
        del self.predictor
        gc.collect()
        torch.cuda.empty_cache()

    @property
    def metadata(self) -> dict[str, Any]:
        """Return the metadata of the model."""
        return {
            "name": self.name,
            "provider": self.provider,
            "torch_dtype": str(self.torch_dtype),
            "config": self.config,
        }

    def set_image_embeddings(
        self,
        image: np.ndarray | "Tensor" | Image,
        image_embedding: Tensor,
        high_resolution_features: Tensor,
    ) -> None:
        """Calculates the image embeddings for the provided image.

        Adapted from https://github.com/facebookresearch/sam2/blob/main/sam2/sam2_image_predictor.py
        (Apache-2.0 License).

        Args:
            image: The input image to embed in RGB format. The image should be in HWC format if np.ndarray, or WHC
                format if PIL Image with or CHW format if torch.Tensor.
            image_embedding: The image embedding tensor.
            high_resolution_features: The high-resolution features tensor.
        """
        with torch.inference_mode():
            self.predictor.reset_predictor()
            # Transform the image to the form expected by the model
            if isinstance(image, np.ndarray):
                self.predictor._orig_hw = [image.shape[:2]]
            elif isinstance(image, Tensor):
                self.predictor._orig_hw = [image.shape[-2:]]
            elif isinstance(image, Image):
                w, h = image.size
                self.predictor._orig_hw = [(h, w)]
            else:
                raise NotImplementedError("Image format not supported")

            self.predictor._features = {
                "image_embed": image_embedding.unsqueeze(0).to(device=self.predictor.model.device),
                "high_res_feats": [
                    features.unsqueeze(0).to(device=self.predictor.model.device)
                    for features in high_resolution_features
                ],
            }
            self.predictor._is_image_set = True

    def image_mask_generation(
        self,
        image: np.ndarray | Image,
        points: list[list[list[int]]] | None,
        labels: list[list[int]] | None,
        boxes: list[list[int]] | None,
        multimask_output: bool = True,
        num_multimask_outputs: int = 3,
        return_image_embedding: bool = False,
        **kwargs: Any,
    ) -> ImageMaskGenerationOutput:
        """Generate masks from the image.

        Args:
            image: Image for the generation.
            points: Points for the mask generation. The first dimension is the number of prompts, the
                second the number of points per mask and the third the coordinates of the points.
            labels: Labels for the mask generation. The first dimension is the number of prompts, the second
                the number of labels per mask.
            boxes: Boxes for the mask generation. The first dimension is the number of prompts, the second
                the coordinates of the boxes.
            multimask_output: Whether to generate multiple masks per prediction.
            num_multimask_outputs: Number of masks to generate per prediction.
            return_image_embedding: Whether to return the image embedding and high-resolution features.
            kwargs: Additional keyword arguments.
        """
        # Check the input list types
        with torch.inference_mode():
            if not isinstance(image, (np.ndarray, Image)):
                raise ValueError("The image should be an numpy array or a PIL image.")
            if (
                points is not None
                and not isinstance(points, list)
                and not all(isinstance(point, list) for point in points)
            ):
                raise ValueError("The points should be a list of lists.")
            if (
                labels is not None
                and not isinstance(labels, list)
                and not all(isinstance(label, list) for label in labels)
            ):
                raise ValueError("The labels should be a list of lists.")
            if boxes is not None and not isinstance(boxes, list):
                if not all(isinstance(box, list) for box in boxes):
                    raise ValueError("The boxes should be a list of lists.")

            if multimask_output and not (num_multimask_outputs == 3):
                raise ValueError("The number of multimask outputs is not configurable for Sam2 and must be 3.")

            # Check the input batch size
            if points is None and labels is not None:
                raise ValueError("Labels are not supported without points.")
            if points is not None and labels is not None:
                if len(points) != len(labels):
                    raise ValueError("The number of points and labels should match.")
            if points is not None and boxes is not None:
                if len(points) != len(boxes):
                    raise ValueError("The number of points and boxes should match.")

            # Check the input shapes and value types
            if points is not None:
                for prompt_point in points:
                    for points_in_mask in prompt_point:
                        if len(points_in_mask) != 2:
                            raise ValueError("Each point should have 2 coordinates.")
                        if not all(isinstance(point, int) for point in points_in_mask):
                            raise ValueError("Each point should be an integer.")
            if labels is not None:
                for i, prompt_label in enumerate(labels):
                    if len(prompt_label) != len(points[i]):  # type: ignore[index]
                        raise ValueError("The number of labels should match the number of points.")

                    if not all(isinstance(label, int) for label in prompt_label):
                        raise ValueError("Each label should be an integer.")
            if boxes is not None:
                for prompt_box in boxes:
                    if len(prompt_box) != 4:
                        raise ValueError("Each box should have 4 coordinates.")
                    if not all(isinstance(box, int) for box in prompt_box):
                        raise ValueError("Each box should be an integer.")

            # Convert inputs to numpy arrays
            if points is not None:
                np_points = [np.array(prompt_points, dtype=np.int32) for prompt_points in points]
            if labels is not None:
                np_labels = [np.array(prompt_labels, dtype=np.int32) for prompt_labels in labels]
            if boxes is not None:
                boxes = np.array(boxes, dtype=np.int32)

            if points is not None:
                # Pad the points and labels to the same length
                # =============================================================================
                # From Hugging Face's implementation (Apache-2.0 License):
                # https://github.com/huggingface/transformers/blob/v4.45.2/src/transformers/models/sam/processing_sam.py#L36
                expected_nb_points = max([point.shape[0] for point in np_points])
                processed_points = []
                for i, point in enumerate(np_points):
                    if point.shape[0] != expected_nb_points:
                        point = np.concatenate(
                            [point, np.zeros((expected_nb_points - point.shape[0], 2)) + -10], axis=0
                        )
                        if labels is not None:
                            np_labels[i] = np.append(np_labels[i], [-10])
                    processed_points.append(point)
                # =============================================================================
                np_points = np.array(processed_points)

            input_points = np_points if points is not None else None
            input_labels = np.array(np_labels) if labels is not None else None

            with torch.autocast(self.predictor.device.type, dtype=self.torch_dtype):
                if not self.predictor._is_image_set:
                    self.predictor.set_image(image)

                masks, scores, _ = self.predictor.predict(
                    point_coords=input_points,
                    point_labels=input_labels,
                    box=boxes,
                    mask_input=None,
                    multimask_output=multimask_output,
                    return_logits=False,
                )

                if len(masks.shape) == 3:
                    masks = np.expand_dims(masks, 0)
                    scores = np.expand_dims(scores, 0)

                if return_image_embedding:
                    image_embedding: Tensor = self.predictor._features["image_embed"]
                    image_embedding_list = image_embedding.to(torch.float32).flatten().tolist()
                    high_resolution_features: list[Tensor] = self.predictor._features["high_res_feats"]
                    high_resolution_features_list = [
                        features.to(torch.float32).flatten().tolist() for features in high_resolution_features
                    ]
                    image_embedding_ndarray = NDArrayFloat(
                        values=image_embedding_list, shape=image_embedding.shape[1:]
                    )
                    high_resolution_features_ndarray = [
                        NDArrayFloat(values=features_list, shape=features.shape[1:])
                        for features, features_list in zip(high_resolution_features, high_resolution_features_list)
                    ]
                else:
                    image_embedding_ndarray = None
                    high_resolution_features_ndarray = None

                masks = ImageMaskGenerationOutput(
                    masks=[
                        [CompressedRLE.from_mask(mask.astype(np.uint8)) for mask in prediction_masks]
                        for prediction_masks in masks
                    ],
                    scores=NDArrayFloat.from_numpy(scores),
                    image_embedding=image_embedding_ndarray,
                    high_resolution_features=high_resolution_features_ndarray,
                )
                return masks

    def video_mask_generation(
        self,
        video_dir: Path,
        objects_ids: list[int],
        frame_indexes: list[int],
        points: list[list[list[int]]] | None = None,
        labels: list[list[int]] | None = None,
        boxes: list[list[int]] | None = None,
        propagate: bool = False,
        **kwargs: Any,
    ) -> VideoMaskGenerationOutput:
        """Generate masks from the video.

        Args:
            video_dir: Directory of the video.
            objects_ids: IDs of the objects to generate masks for.
            frame_indexes: Indexes of the frames where the objects are located.
            points: Points for the mask generation. The first fimension is the number of objects, the
                second the number of points for each object and the third the coordinates of the points.
            labels: Labels for the mask generation. The first fimension is the number of objects, the second
                the number of labels for each object.
            boxes: Boxes for the mask generation. The first fimension is the number of objects, the second
                the coordinates of the boxes.
            propagate: Whether to propagate the masks in the video.
            kwargs: Additional keyword arguments.

        Returns:
            Output of the generation.
        """
        # Check the input list types
        with torch.inference_mode():
            if not isinstance(video_dir, (Path)) or not video_dir.exists():
                raise ValueError("The video_dir should be a valid path.")
            if (
                points is not None
                and not isinstance(points, list)
                and not all(isinstance(point, list) for point in points)
            ):
                raise ValueError("The points should be a list of lists.")
            if (
                labels is not None
                and not isinstance(labels, list)
                and not all(isinstance(label, list) for label in labels)
            ):
                raise ValueError("The labels should be a list of lists.")
            if boxes is not None and not isinstance(boxes, list):
                if not all(isinstance(box, list) for box in boxes):
                    raise ValueError("The boxes should be a list of lists.")

            # Check the input batch size
            if points is None and labels is not None:
                raise ValueError("Labels are not supported without points.")
            if points is not None and labels is not None:
                if len(points) != len(labels):
                    raise ValueError("The number of points and labels should match.")
            if points is not None and boxes is not None:
                if len(points) != len(boxes):
                    raise ValueError("The number of points and boxes should match.")

            # Check the input shapes and value types
            if points is not None:
                for prompt_point in points:
                    for points_in_mask in prompt_point:
                        if len(points_in_mask) != 2:
                            raise ValueError("Each point should have 2 coordinates.")
                        if not all(isinstance(point, int) for point in points_in_mask):
                            raise ValueError("Each point should be an integer.")
            if labels is not None:
                for i, prompt_label in enumerate(labels):
                    if len(prompt_label) != len(points[i]):  # type: ignore[index]
                        raise ValueError("The number of labels should match the number of points.")

                    if not all(isinstance(label, int) for label in prompt_label):
                        raise ValueError("Each label should be an integer.")
            if boxes is not None:
                for prompt_box in boxes:
                    if len(prompt_box) != 4:
                        raise ValueError("Each box should have 4 coordinates.")
                    if not all(isinstance(box, int) for box in prompt_box):
                        raise ValueError("Each box should be an integer.")

            # Convert inputs to numpy arrays
            if points is not None:
                input_points = [np.array(prompt_points, dtype=np.int32) for prompt_points in points]
            else:
                input_points = [None for _ in objects_ids]
            if labels is not None:
                input_labels = [np.array(prompt_labels, dtype=np.int32) for prompt_labels in labels]
            else:
                input_labels = [None for _ in objects_ids]
            if boxes is not None:
                input_boxes = [np.array(box, dtype=np.int32) for box in boxes]
            else:
                input_boxes = [None for _ in objects_ids]

            video_segments: dict[int, dict[int, np.ndarray]] = {}
            with torch.autocast(self.predictor.device.type, dtype=self.torch_dtype):
                inference_state = self.predictor.init_state(video_path=str(video_dir))
                for object_id, object_frame_index, object_points, object_labels, object_boxes in zip(
                    objects_ids, frame_indexes, input_points, input_labels, input_boxes
                ):
                    _, out_obj_ids, out_mask_logits = self.predictor.add_new_points_or_box(
                        inference_state=inference_state,
                        frame_idx=object_frame_index,
                        obj_id=object_id,
                        points=object_points,
                        labels=object_labels,
                        box=object_boxes,
                    )

                    if not propagate:
                        video_segments[object_frame_index] = {
                            out_obj_id: (out_mask_logits[i] > 0.0).cpu().numpy()
                            for i, out_obj_id in enumerate(out_obj_ids)
                        }

                if propagate:
                    video_segments = {}  # video_segments contains the per-frame segmentation results
                    for out_frame_idx, out_obj_ids, out_mask_logits in self.predictor.propagate_in_video(
                        inference_state
                    ):
                        video_segments[out_frame_idx] = {
                            out_obj_id: (out_mask_logits[i] > 0.0).cpu().numpy()
                            for i, out_obj_id in enumerate(out_obj_ids)
                        }

            out_objects_ids = []
            out_frame_indexes = []
            out_masks = []

            for frame_index, object_masks in video_segments.items():
                for object_id, mask in object_masks.items():
                    out_objects_ids.append(object_id)
                    out_frame_indexes.append(frame_index)
                    out_masks.append(mask)

            return VideoMaskGenerationOutput(
                objects_ids=out_objects_ids,
                frame_indexes=out_frame_indexes,
                masks=[CompressedRLE.from_mask(mask.astype(np.uint8)) for mask in out_masks],
            )



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Multimodal tasks."""

from .task import Task


class MultimodalImageNLPTask(Task):
    """Multimodal tasks."""

    CAPTIONING = "image_captioning"
    CONDITIONAL_GENERATION = "text_image_conditional_generation"
    EMBEDDING = "text_image_embedding"
    MATCHING = "text_image_matching"



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

from .fixtures.client import simple_pixano_inference_client



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Celery configuration file."""

import logging
import os
import signal
import subprocess
import sys
import time
from datetime import datetime
from subprocess import Popen
from typing import Any, cast

from celery import Celery
from celery.result import AsyncResult
from fastapi.encoders import jsonable_encoder

from pixano_inference.models import BaseInferenceModel
from pixano_inference.providers.base import BaseProvider, ModelProvider
from pixano_inference.providers.utils import instantiate_provider
from pixano_inference.pydantic import (
    CeleryTask,
    ImageMaskGenerationRequest,
    ImageZeroShotDetectionRequest,
    ModelConfig,
    TextImageConditionalGenerationRequest,
    VideoMaskGenerationRequest,
)
from pixano_inference.tasks import ImageTask, MultimodalImageNLPTask, Task, VideoTask, str_to_task
from pixano_inference.utils.package import assert_torch_installed, is_torch_installed


if is_torch_installed():
    import torch

uvicorn_logger = logging.getLogger("uvicorn.error")

queues_to_workers: dict[str, Popen] = {}
worker_model: BaseInferenceModel
worker_provider: BaseProvider
worker_task: Task
is_model_initialized: bool = False


def create_celery_app() -> Celery:
    """Create a new celery app."""
    redis_url: str = os.environ.get("REDIS_URL", "localhost")
    redis_port: int = int(os.environ.get("REDIS_PORT", 6379))
    redis_db_number: int = int(os.environ.get("REDIS_DB_NUMBER", 0))

    redis_url = f"redis://{redis_url}:{redis_port}/{redis_db_number}"

    app = Celery(__name__, broker=redis_url, backend=redis_url)

    return app


celery_app = create_celery_app()


@celery_app.task
def instantiate_model(provider: str, model_config: dict[str, Any], gpu: int | None) -> None:
    """Instantiate a model."""
    global worker_provider, worker_model, worker_task, is_model_initialized

    if is_model_initialized:
        raise ValueError("Do not initialize twice a model.")
    is_model_initialized = True

    assert_torch_installed()
    device = torch.device(f"cuda:{gpu}") if gpu is not None else torch.device("cpu")
    worker_provider = instantiate_provider(provider)
    worker_provider = cast(ModelProvider, worker_provider)
    worker_model = worker_provider.load_model(**model_config, device=device)
    worker_task = str_to_task(model_config["task"])


@celery_app.task
def delete_model() -> None:
    """Delete model."""
    global worker_model

    try:
        worker_model.delete()
    except NameError:
        pass


@celery_app.task
def predict(request: dict[str, Any]) -> dict[str, Any]:
    """Run a model inference from the request."""
    global worker_provider, worker_model, worker_task

    start_time = time.time()
    match worker_task:
        case ImageTask.MASK_GENERATION:
            output = worker_provider.image_mask_generation(
                request=ImageMaskGenerationRequest.model_construct(request), model=worker_model
            )
        case MultimodalImageNLPTask.CONDITIONAL_GENERATION:
            output = worker_provider.text_image_conditional_generation(
                request=TextImageConditionalGenerationRequest.model_construct(**request), model=worker_model
            )
        case VideoTask.MASK_GENERATION:
            output = worker_provider.video_mask_generation(
                request=VideoMaskGenerationRequest.model_construct(**request), model=worker_model
            )
        case ImageTask.ZERO_SHOT_DETECTION:
            output = worker_provider.image_zero_shot_detection(
                request=ImageZeroShotDetectionRequest.model_construct(**request), model=worker_model
            )
        case _:
            raise ValueError(f"Unknown task: {worker_task}")
    response = {
        "timestamp": datetime.now(),
        "processing_time": time.time() - start_time,
        "metadata": worker_model.metadata,
        "data": output.model_dump(),
    }
    return response


def model_queue_name(model_name: str) -> str:
    """Get the name of the queue for a given model."""
    return f"{model_name}_queue"


def add_celery_worker_and_queue(provider: str, model_config: ModelConfig, gpu: int | None) -> CeleryTask:
    """Add a new worker and a queue to the celery app to handle model."""
    queue = model_queue_name(model_config.name)
    celery_app.control.add_consumer(queue=model_queue_name(model_config.name), reply=True)

    command = [
        sys.executable,
        "-m",
        "celery",
        "-A",
        "pixano_inference.celery.celery_app",
        "worker",
        "--loglevel=INFO",
        "-Q",
        queue,
        "--pool=solo",
    ]
    worker = Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        start_new_session=True,
    )

    uvicorn_logger.info(f"Spawned Celery worker {worker.pid} handling model {model_config.name}.")
    queues_to_workers[queue] = worker

    task: AsyncResult = instantiate_model.apply_async(
        (provider, jsonable_encoder(model_config.model_dump()), gpu), queue=queue
    )
    task_result, result = list(task.collect())[0]

    return CeleryTask(id=task.id, status=task_result.status)


def delete_celery_worker_and_queue(model_name: str):
    """Delete a worker and a queue of the celery app that handled a model."""
    queue = model_queue_name(model_name)

    command = [
        sys.executable,
        "-m",
        "celery",
        "-Q",
        queue,
        "purge",
    ]
    purge_process = Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        start_new_session=True,
    )
    purge_process.wait()
    purge_process.kill()
    try:
        worker = queues_to_workers.pop(queue)
    except KeyError:  # Instantiation failed before storing the worker
        pass
    else:
        delete_model.apply_async(queue=queue).get()
        os.killpg(os.getpgid(worker.pid), signal.SIGTERM)
        worker.wait()
        uvicorn_logger.info(f"Killed Celery worker {worker.pid} handling model {model_name}.")
    celery_app.control.cancel_consumer(queue=queue)



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""URL utils."""

import re


# TODO: improve this regex pattern
url_validation_regex = r"^(file://|https?://|s3://)[^\s]+$"


def is_url(url: str) -> bool:
    """Check if a string is a valid URL."""
    return re.match(url_validation_regex, url) is not None



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Pydantic base models for request and response."""

from abc import ABC
from datetime import datetime
from typing import Any, TypeVar

from pydantic import BaseModel


T = TypeVar("T", bound=BaseModel)


class BaseRequest(BaseModel, ABC):
    """Base request model.

    Attributes:
        model: Name of the model.
    """

    model: str

    def to_base_model(self, base_model: type[T]) -> T:
        """Convert request to input type."""
        if not issubclass(base_model, BaseModel):
            raise ValueError(f"base_model must be a subclass of pydantic's BaseModel, got {base_model.__name__}.")
        return base_model.model_validate(self.model_dump(include=list(base_model.model_fields.keys())))


class APIRequest(BaseRequest, ABC):
    """API request model.

    Attributes:
        api_key: API key.
        secret_key: Secret key.
    """

    api_key: str
    secret_key: str


class CeleryTask(BaseModel):
    """Celery task model.

    Attributes:
        task_id: ID of the celery task.
        status: Status of the celery task.
    """

    id: str
    status: str


class BaseResponse(BaseModel, ABC):
    """Base response model.

    Attributes:
        task_id: ID of the celery task.
        status: Status of the celery task.
        timestamp: Timestamp of the response.
        processing_time: Processing time of the response.
        metadata: Metadata of the response.
        data: Data of the response.
    """

    id: str
    status: str
    timestamp: datetime
    processing_time: float = 0.0
    metadata: dict[str, Any]
    data: Any



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Pydantic models for N-dimensional arrays."""

from abc import ABC
from typing import TYPE_CHECKING, ClassVar, Generic, TypeVar

import numpy as np
from pydantic import BaseModel, field_validator
from typing_extensions import Self

from pixano_inference.utils.package import assert_torch_installed, is_torch_installed


if is_torch_installed():
    import torch

if TYPE_CHECKING:
    from torch import Tensor

T = TypeVar("T")


class NDArray(BaseModel, Generic[T], ABC):
    """Represents an N-dimensional array.

    Attributes:
        values: The list of values.
        shape: The shape of the array, represented as a list of integers.
        np_dtype: The NumPy data type of the array.
    """

    values: list[T]
    shape: list[int]
    np_dtype: ClassVar[np.dtype]

    @field_validator("shape", mode="after")
    @classmethod
    def _validate_shape(cls, v: list[int]) -> list[int]:
        if len(v) < 1:
            raise ValueError("Shape must have at least one element.")
        elif any(s < 1 for s in v):
            raise ValueError("Shape elements must be positive.")
        return v

    @classmethod
    def from_numpy(cls, arr: np.ndarray) -> Self:
        """Create an instance of the class from a NumPy array.

        Args:
            arr: The NumPy array to convert.

        Returns:
            An instance of the class with values and shape derived from
                the input array.
        """
        shape = list(arr.shape)
        arr = arr.astype(dtype=cls.np_dtype)
        return cls(
            values=arr.reshape(-1).tolist(),
            shape=shape,
        )

    @classmethod
    def from_torch(cls, tensor: "Tensor") -> Self:
        """Create an instance of the class from a PyTorch tensor.

        Args:
            tensor: The PyTorch tensor to convert.

        Returns:
            An instance of the class with values and shape derived from
                the input tensor.
        """
        assert_torch_installed()
        return cls.from_numpy(tensor.cpu().numpy())

    def to_numpy(self) -> np.ndarray:
        """Convert the instance to a NumPy array.

        Returns:
            A NumPy array with values and shape derived from the instance.
        """
        array = np.array(self.values, dtype=self.np_dtype).reshape(self.shape)
        return array

    def to_torch(self) -> "Tensor":
        """Convert the instance to a PyTorch tensor.

        Returns:
            A PyTorch tensor with values and shape derived from the instance.
        """
        assert_torch_installed()
        return torch.from_numpy(self.to_numpy())


class NDArrayFloat(NDArray[float]):
    """Represents an N-dimensional array of 32-bit floating-point values.

    Attributes:
        values: The list of 32-bit floating-point values in the array.
        shape: The shape of the array, represented as a list of integers.
        np_dtype: The NumPy data type of the array.
    """

    np_dtype: ClassVar[np.dtype] = np.float32



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Pydantic models for text-image conditional generation task."""

from pathlib import Path
from typing import Any

from pydantic import BaseModel

from pixano_inference.pydantic.base import BaseRequest, BaseResponse


class TextImageConditionalGenerationInput(BaseModel):
    """Input for text-image conditional generation.

    Attributes:
        prompt: Prompt for the generation. Can be a string or a list of dictionaries to apply a chat template.
        images: Images for the generation. Can be None if images are passed in the prompt.
        max_new_tokens: Maximum number of new tokens to generate.
        temperature: Temperature for the generation.
    """

    prompt: str | list[dict[str, Any]]
    images: list[str | Path] | None = None
    max_new_tokens: int
    temperature: float = 1.0


class TextImageConditionalGenerationRequest(BaseRequest, TextImageConditionalGenerationInput):
    """Request for text-image conditional generation."""

    def to_input(self) -> TextImageConditionalGenerationInput:
        """Convert the request to the input."""
        return self.to_base_model(TextImageConditionalGenerationInput)


class UsageConditionalGeneration(BaseModel):
    """Usage metadata of the model for text-image conditional generation.

    Attributes:
        prompt_tokens: Number of tokens in the prompt.
        completion_tokens: Number of tokens in the completion.
        total_tokens: Total number of tokens.
    """

    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class TextImageConditionalGenerationOutput(BaseModel):
    """Output for text-image conditional generation.

    Attributes:
        generated_text: Generated text.
        usage: Usage of the model for the generation.
        generation_config: Configuration for the generation.
    """

    generated_text: str
    usage: UsageConditionalGeneration
    generation_config: dict[str, Any] = {}


class TextImageConditionalGenerationResponse(BaseResponse):
    """Response for text-image conditional generation.

    Attributes:
        data: Output of the generation.
    """

    data: TextImageConditionalGenerationOutput



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""API routes for the app."""

from typing import Annotated

from fastapi import APIRouter, Depends

from pixano_inference.pydantic.models import ModelInfo
from pixano_inference.settings import Settings, get_pixano_inference_settings


router = APIRouter(prefix="/app", tags=["App"])


@router.get("/settings/", response_model=Settings)
async def get_settings(
    settings: Annotated[Settings, Depends(get_pixano_inference_settings)],
) -> Settings:
    """Get the current settings of the app."""
    return settings


@router.get("/models/", response_model=list[ModelInfo])
async def get_list_models(
    settings: Annotated[Settings, Depends(get_pixano_inference_settings)],
) -> list[ModelInfo]:
    """List all models available in the app."""
    models = [ModelInfo(name=model_name, task=task) for model_name, task in settings.models_to_task.items()]
    return models



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Video tasks."""

from .task import Task


class VideoTask(Task):
    """Video tasks."""

    MASK_GENERATION = "video_mask_generation"



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Pydantic models for image zero shot detection task."""

from pathlib import Path

from pydantic import BaseModel, ConfigDict

from pixano_inference.pydantic.base import BaseRequest, BaseResponse


class ImageZeroShotDetectionInput(BaseModel):
    """Input for image zero shot detection.

    Attributes:
        image: Image for image zero shot detection.
        classes: List of classes to detect.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    image: str | Path
    classes: str | list[str]
    box_threshold: float = 0.5
    text_threshold: float = 0.5


class ImageZeroShotDetectionRequest(BaseRequest, ImageZeroShotDetectionInput):
    """Request for image zero shot detection."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def to_input(self) -> ImageZeroShotDetectionInput:
        """Convert the request to the input."""
        return self.to_base_model(ImageZeroShotDetectionInput)


class ImageZeroShotDetectionOutput(BaseModel):
    """Output for image zero shot detection.

    Attributes:
        boxes: List of boxes detected by the model.
        scores: List of scores detected by the model (higher is better).
        classes: List of class names associated with each box and score.
    """

    boxes: list[list[int]]
    scores: list[float]
    classes: list[str]


class ImageZeroShotDetectionResponse(BaseResponse):
    """Response for image zero shot detection.

    Attributes:
        data: Output of the generation.
    """

    data: ImageZeroShotDetectionOutput



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

# ================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# ================================

"""Check if all files contain the required header."""

from pathlib import Path


ROOT = Path(__file__).parent

# Headers for each file type
HEADERS = {
    ".py": """# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================
\n""",
    ".yml": """# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================
\n""",
    ".yaml": """# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================
\n""",
    ".md": """<!---
# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================
--->
\n""",
}

# Files to exclude
EXCLUDE_FILES: set[str] = {
    "PULL_REQUEST_TEMPLATE.md",
    "bug_report.md",
    "feature_request.md",
}

# Files to exclude
EXCLUDE_PATHS: set[str] = set()


def check_header(file_path: Path, header: str):
    """Check if the file contains the required header."""
    with file_path.open("r", encoding="utf-8") as file:
        content = file.read()
        return content.startswith(header) or content == header[:-1]


def main():
    """Check if all files contain the required headers."""
    missing_headers = []
    for file_path in ROOT.rglob("*"):
        if not any(part.name in EXCLUDE_PATHS for part in file_path.parents):
            ext = file_path.suffix
            if ext in HEADERS and file_path.name not in EXCLUDE_FILES:
                if not check_header(file_path, HEADERS[ext]):
                    missing_headers.append(str(file_path))

    if missing_headers:
        exit(Exception("The following files are missing the required headers:\n" + "\n- ".join(missing_headers)))


if __name__ == "__main__":
    main()



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Natural Language Processing tasks."""

from .task import Task


class NLPTask(Task):
    """Natural Language Processing tasks."""

    CAUSAL_LM = "causal_lm"
    CONDITONAL_GENERATION = "text_conditional_generation"
    MASKED_LM = "masked_lm"
    MASK_GENERATION = "text_mask_generation"
    MULTIPLE_CHOICE = "multiple_choice"
    NEXT_SENTENCE_PREDICTION = "next_sentence_prediction"
    QUESTION_ANSWERING = "question_answering"
    SEQUENCE_CLASSIFICATION = "sequence_classification"
    TEXT_ENCODING = "text_encoding"
    TOKEN_CLASSIFICATION = "token_classification"



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Base class for inference models."""

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any

from pixano_inference.pydantic import (
    ImageMaskGenerationOutput,
    ImageZeroShotDetectionOutput,
    TextImageConditionalGenerationOutput,
    VideoMaskGenerationOutput,
)


class ModelStatus(Enum):
    """Current status of the model.

    Attributes:
    - IDLE: waiting for an input.
    - RUNNING: computing.
    """

    IDLE = 0
    RUNNING = 1


class BaseInferenceModel(ABC):
    """Base class for inference models."""

    def __init__(self, name: str, provider: str):
        """Initialize the model.

        Args:
            name: Name of the model.
            provider: Provider of the model.
        """
        self.name = name
        self.provider = provider
        self._status = ModelStatus.IDLE

    @property
    def status(self) -> ModelStatus:
        """Get the status of the model."""
        return self._status

    @status.setter
    def status(self, new_status: ModelStatus):
        """Set the status of the model.

        It should be handled outside of the Inference Model by a controller to handle requests sequentially.
        """
        if not isinstance(new_status, ModelStatus):
            raise ValueError(f"Status should be a ModelStatus, got {new_status}.")
        self._status = new_status

    @property
    @abstractmethod
    def metadata(self) -> dict[str, Any]:
        """Return the metadata of the model."""
        ...

    @abstractmethod
    def delete(self):
        """Delete the model."""
        ...

    def image_mask_generation(self, *args: Any, **kwargs) -> ImageMaskGenerationOutput:
        """Generate a mask from the image."""
        raise NotImplementedError("This model does not support image mask generation.")

    def text_image_conditional_generation(self, *args: Any, **kwargs) -> TextImageConditionalGenerationOutput:
        """Generate text from an image and a prompt."""
        raise NotImplementedError("This model does not support text-image conditional generation.")

    def video_mask_generation(self, *args: Any, **kwargs) -> VideoMaskGenerationOutput:
        """Generate a mask from the video."""
        raise NotImplementedError("This model does not support video mask generation.")

    def image_zero_shot_detection(self, *args: Any, **kwargs) -> ImageZeroShotDetectionOutput:
        """Perform zero shot detection on an image."""
        raise NotImplementedError("This model does not support image zero shot detection.")



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Main application entry point."""

import click
import uvicorn
from fastapi import APIRouter, FastAPI
from fastapi.responses import RedirectResponse

from . import routers


router = APIRouter()


@router.get("/", include_in_schema=False)
async def docs_redirect() -> RedirectResponse:
    """Redirect homepage to docs."""
    return RedirectResponse(url="/docs")


def create_app() -> FastAPI:
    """Create the FastAPI application."""
    app = FastAPI()

    app.include_router(routers.tasks.router)
    app.include_router(routers.providers.router)
    app.include_router(routers.app.router)

    app.include_router(router)
    return app


fast_api_app = create_app()


@click.command(context_settings={"auto_envvar_prefix": "UVICORN"})
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="Pixano Inference app URL host",
    show_default=True,
)
@click.option(
    "--port",
    type=int,
    default=80,
    help="Pixano Inference app URL port",
    show_default=True,
)
def serve(host: str, port: int):
    """Main application entry point."""
    config = uvicorn.Config(fast_api_app, host=host, port=port)
    server = uvicorn.Server(config)
    server.run()



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Provider for Hugging Face Transformers models."""

from pathlib import Path
from typing import Any

from PIL.Image import Image

from pixano_inference.models.transformers import TransformerModel
from pixano_inference.providers.registry import register_provider
from pixano_inference.pydantic import (
    ImageMaskGenerationOutput,
    ImageMaskGenerationRequest,
    ImageZeroShotDetectionOutput,
    ImageZeroShotDetectionRequest,
    TextImageConditionalGenerationOutput,
    TextImageConditionalGenerationRequest,
)
from pixano_inference.tasks import ImageTask, MultimodalImageNLPTask, NLPTask, Task, str_to_task
from pixano_inference.utils.image import convert_string_to_image
from pixano_inference.utils.package import (
    assert_transformers_installed,
    is_torch_installed,
    is_transformers_installed,
)
from pixano_inference.utils.vector import vector_to_tensor

from .base import ModelProvider


if is_torch_installed():
    import torch

if is_transformers_installed():
    from transformers import AutoProcessor, BitsAndBytesConfig
    from transformers.modeling_utils import PreTrainedModel


def get_transformer_automodel_from_pretrained(
    pretrained_model_name_or_path: str | Path, task: Task, **model_kwargs: Any
):
    """Get a transformer model from transformers using automodel.

    Args:
        pretrained_model_name_or_path: Name or path of the pretrained model.
        task: Task of the model.
        model_kwargs: Additional keyword arguments for the model.
    """
    assert_transformers_installed()
    pretrained_model_name_or_path = str(pretrained_model_name_or_path)
    if isinstance(task, ImageTask):
        match task:
            case ImageTask.CLASSIFICATION:
                from transformers import AutoModelForImageClassification

                return AutoModelForImageClassification.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case ImageTask.DEPTH_ESTIMATION:
                from transformers import AutoModelForDepthEstimation

                return AutoModelForDepthEstimation.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case ImageTask.FEATURE_EXTRACTION:
                from transformers import AutoModel

                return AutoModel.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case ImageTask.KEYPOINT_DETECTION:
                from transformers import AutoModelForKeypointDetection

                return AutoModelForKeypointDetection.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case ImageTask.MASK_GENERATION:
                from transformers import AutoModelForMaskGeneration

                return AutoModelForMaskGeneration.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case ImageTask.OBJECT_DETECTION:
                from transformers import AutoModelForObjectDetection

                return AutoModelForObjectDetection.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case ImageTask.SEMANTIC_SEGMENTATION:
                from transformers import AutoModelForSemanticSegmentation

                return AutoModelForSemanticSegmentation.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case ImageTask.INSTANCE_SEGMENTATION:
                from transformers import AutoModelForInstanceSegmentation

                return AutoModelForInstanceSegmentation.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case ImageTask.UNIVERSAL_SEGMENTATION:
                from transformers import AutoModelForUniversalSegmentation

                return AutoModelForUniversalSegmentation.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case ImageTask.ZERO_SHOT_CLASSIFICATION:
                from transformers import AutoModelForZeroShotImageClassification

                return AutoModelForZeroShotImageClassification.from_pretrained(
                    pretrained_model_name_or_path, **model_kwargs
                )
            case ImageTask.ZERO_SHOT_DETECTION:
                from transformers import AutoModelForZeroShotObjectDetection

                return AutoModelForZeroShotObjectDetection.from_pretrained(
                    pretrained_model_name_or_path, **model_kwargs
                )
    elif isinstance(task, NLPTask):
        match task:
            case NLPTask.CAUSAL_LM:
                from transformers import AutoModelForCausalLM

                return AutoModelForCausalLM.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case NLPTask.MASKED_LM:
                from transformers import AutoModelForMaskedLM

                return AutoModelForMaskedLM.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case NLPTask.MASK_GENERATION:
                from transformers import AutoModelForMaskGeneration

                return AutoModelForMaskGeneration.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case NLPTask.SEQUENCE_CLASSIFICATION:
                from transformers import AutoModelForSequenceClassification

                return AutoModelForSequenceClassification.from_pretrained(
                    pretrained_model_name_or_path, **model_kwargs
                )
            case NLPTask.MULTIPLE_CHOICE:
                from transformers import AutoModelForMultipleChoice

                return AutoModelForMultipleChoice.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case NLPTask.NEXT_SENTENCE_PREDICTION:
                from transformers import AutoModelForNextSentencePrediction

                return AutoModelForNextSentencePrediction.from_pretrained(
                    pretrained_model_name_or_path, **model_kwargs
                )
            case NLPTask.TOKEN_CLASSIFICATION:
                from transformers import AutoModelForTokenClassification

                return AutoModelForTokenClassification.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case NLPTask.QUESTION_ANSWERING:
                from transformers import AutoModelForQuestionAnswering

                return AutoModelForQuestionAnswering.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case NLPTask.TEXT_ENCODING:
                from transformers import AutoModelForTextEncoding

                return AutoModelForTextEncoding.from_pretrained(pretrained_model_name_or_path, **model_kwargs)
            case _:
                raise ValueError(f"Task {task} not supported.")


def get_conditional_generation_transformer_from_pretrained(
    name: str, path: Path | str | None, **model_kwargs: Any
) -> PreTrainedModel:
    """Get a transformer model from transformers using automodel.

    Args:
        name: Name of the model.
        path: Path to the model or its Hugging Face hub's identifier.
        model_kwargs: Additional keyword arguments for the model.

    Returns:
        Model from Transformers.
    """
    name = name.lower()
    if "llava" in name:
        if "next" in name:
            if "video" in name:
                from transformers import LlavaNextVideoForConditionalGeneration

                model = LlavaNextVideoForConditionalGeneration.from_pretrained(path, **model_kwargs)
            else:
                from transformers import LlavaNextForConditionalGeneration

                model = LlavaNextForConditionalGeneration.from_pretrained(path, **model_kwargs)
        else:
            from transformers import LlavaForConditionalGeneration

            model = LlavaForConditionalGeneration.from_pretrained(path, **model_kwargs)
    else:
        raise ValueError(f"Model {name} not supported.")
    return model


@register_provider("transformers")
class TransformersProvider(ModelProvider):
    """Provider for Hugging Face Transformers models."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the transformer provider."""
        assert_transformers_installed()
        super().__init__(*args, **kwargs)

    def load_model(
        self,
        name: str,
        task: Task | str,
        device: "torch.device",
        path: Path | str | None = None,
        processor_config: dict = {},
        config: dict = {},
    ) -> TransformerModel:
        """Load a model from transformers.

        Args:
            name: Name of the model.
            task: Task of the model.
            device: Device to use for the model.
            path: Path to the model or its Hugging Face hub's identifier.
            processor_config: Configuration for the processor.
            config: Configuration for the model.

        Returns:
            Loaded model.
        """
        if path is None:
            raise ValueError("Path is required to load a model from transformers.")
        if isinstance(task, str):
            task = str_to_task(task)
        processor = AutoProcessor.from_pretrained(path, **processor_config)

        if (quantization_config := config.pop("quantization_config", None)) is not None:
            quantization_config = BitsAndBytesConfig(**quantization_config)
            config["quantization_config"] = quantization_config

        model = get_transformer_automodel_from_pretrained(path, task, device_map=device, **config)
        if model is None:
            if task in [NLPTask.CONDITONAL_GENERATION, MultimodalImageNLPTask.CONDITIONAL_GENERATION]:
                model = get_conditional_generation_transformer_from_pretrained(name, path, device_map=device, **config)

        model = model.eval()
        model = torch.compile(model)

        our_model = TransformerModel(name, path, processor, model)
        return our_model

    def image_mask_generation(
        self,
        request: ImageMaskGenerationRequest,
        model: TransformerModel,  # type: ignore[override]
        *args: Any,
        **kwargs: Any,
    ) -> ImageMaskGenerationOutput:
        """Generate a mask from the image.

        Args:
            request: Request for the generation.
            model: Model to use for the generation.
            args: Additional arguments.
            kwargs: Additional keyword arguments.

        Returns:
            Output of the generation
        """
        request_input = request.to_input()
        image = convert_string_to_image(request_input.image)

        if request_input.image_embedding is not None:
            image_embedding = vector_to_tensor(request_input.image_embedding)

        model_input = request_input.model_dump(exclude=["image", "image_embedding"])
        model_input["image"] = image
        model_input["image_embedding"] = image_embedding if request_input.image_embedding is not None else None
        output = model.image_mask_generation(**model_input)
        return output

    def text_image_conditional_generation(
        self,
        request: TextImageConditionalGenerationRequest,
        model: TransformerModel,  # type: ignore[override]
        *args: Any,
        **kwargs: Any,
    ) -> TextImageConditionalGenerationOutput:
        """Generate text from an image and a prompt.

        Args:
            request: Request for text-image conditional generation.
            model: Model for text-image conditional generation
            args: Additional arguments.
            kwargs: Additional keyword arguments.

        Returns:
            Output of text-image conditional generation.
        """
        model_input = request.to_input()

        images: list[Image] | None
        if model_input.images is None:
            if isinstance(model_input.prompt, str):
                raise ValueError("Images must be provided if the prompt is a string.")
            images = []
            for message in model_input.prompt:
                new_content = []
                for content in message["content"]:
                    if content["type"] == "image_url":
                        images.append(convert_string_to_image(content["image_url"]["url"]))
                        new_content.append({"type": "image"})
                    else:
                        new_content.append(content)
                message["content"] = new_content

        else:
            images = (
                [convert_string_to_image(image) for image in model_input["images"]]
                if len(model_input["images"]) > 0
                else None
            )

        model_input_dump = model_input.model_dump()
        model_input_dump["images"] = images
        output = model.text_image_conditional_generation(**model_input_dump)
        return output

    def image_zero_shot_detection(
        self,
        request: ImageZeroShotDetectionRequest,
        model: TransformerModel,  # type: ignore[override]
        *args: Any,
        **kwargs: Any,
    ) -> ImageZeroShotDetectionOutput:
        """Perform zero-shot image detection."""
        request_input = request.to_input()

        image = convert_string_to_image(request_input.image)
        classes = request.classes
        if isinstance(classes, list):
            classes = ". ".join(classes)

        model_input = request_input.model_dump(exclude=["image", "classes"])
        model_input["image"] = image
        model_input["classes"] = classes

        output = model.image_zero_shot_detection(**model_input)
        return output



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Utils for the routers."""

from celery import states
from celery.result import AsyncResult
from fastapi import HTTPException
from fastapi.encoders import jsonable_encoder

from pixano_inference.celery import celery_app, model_queue_name, predict
from pixano_inference.pydantic import BaseRequest, BaseResponse, CeleryTask
from pixano_inference.settings import Settings
from pixano_inference.tasks import Task


async def execute_task_request(request: BaseRequest, task: Task, settings: Settings) -> CeleryTask:
    """Execute a task request.

    Args:
        request: Request to execute.
        task: Task to execute
        settings: Settings of the app.

    Returns:
        Response of the request
    """
    model_name = request.model
    if model_name not in settings.models:
        raise HTTPException(404, detail=f"Model {model_name} is not registered.")
    model_task = settings.models_to_task[model_name]
    if task.value != model_task:
        raise HTTPException(400, detail=f"Model {model_name} does not support the {task.value} task.")

    queue = model_queue_name(model_name)
    celery_task: AsyncResult = predict.apply_async((jsonable_encoder(request),), queue=queue)
    return CeleryTask(id=celery_task.id, status=states.PENDING)


async def get_task_result(task_id: str, response_type: type[BaseResponse]) -> CeleryTask | BaseResponse:
    """Get the result of a task.

    Args:
        task_id: ID of the task to retrieve.
        response_type: Type of response to return.

    Returns:
        Response for the task generation.
    """
    task_result = celery_app.AsyncResult(task_id)
    status, result = task_result.status, task_result.result
    if status in [states.PENDING, states.REVOKED, states.STARTED, states.FAILURE, states.REVOKED]:
        return CeleryTask(id=task_id, status=status)
    elif status != states.SUCCESS:
        raise HTTPException(status_code=500, detail=f"Unknown task status {status}")
    result["id"] = task_id
    result["status"] = status
    response = response_type.model_construct(**result)
    task_result.forget()
    return response


async def delete_task(task_id: str) -> None:
    """Delete a task."""
    celery_app.control.revoke(task_id, terminate=True)
    return



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

# ruff: noqa: F401
# ruff: noqa: D104

from .image import (
    CompressedRLE,
    ImageMaskGenerationInput,
    ImageMaskGenerationOutput,
    ImageMaskGenerationRequest,
    ImageMaskGenerationResponse,
    ImageZeroShotDetectionInput,
    ImageZeroShotDetectionOutput,
    ImageZeroShotDetectionRequest,
    ImageZeroShotDetectionResponse,
)
from .multimodal import (
    TextImageConditionalGenerationInput,
    TextImageConditionalGenerationOutput,
    TextImageConditionalGenerationRequest,
    TextImageConditionalGenerationResponse,
    UsageConditionalGeneration,
)
from .video import (
    VideoMaskGenerationInput,
    VideoMaskGenerationOutput,
    VideoMaskGenerationRequest,
    VideoMaskGenerationResponse,
)



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Pixano inference client."""

import asyncio
from typing import Any, Literal, overload

import httpx
import requests  # type: ignore[import-untyped]
from celery import states
from fastapi import HTTPException
from fastapi.encoders import jsonable_encoder
from httpx import Response
from pydantic import field_validator

from .pydantic import (
    BaseRequest,
    BaseResponse,
    CeleryTask,
    ImageMaskGenerationRequest,
    ImageMaskGenerationResponse,
    ImageZeroShotDetectionRequest,
    ImageZeroShotDetectionResponse,
    ModelConfig,
    ModelInfo,
    TextImageConditionalGenerationRequest,
    TextImageConditionalGenerationResponse,
    VideoMaskGenerationRequest,
    VideoMaskGenerationResponse,
)
from .settings import Settings
from .utils import is_url


class InferenceTooLongError(Exception):
    """Exeption when inference took too long."""

    pass


def _validate_task_id_asynchronous(task_id: str | None, asynchronous: bool):
    if not asynchronous and task_id is not None:
        raise ValueError("Task id must be None for synchronous calls.")


def _validate_poll_interval_timeout(poll_interval: float, timeout: float):
    if not isinstance(poll_interval, (float, int)):
        raise ValueError("Poll interval should be a number that define an interval in seconds.")
    if not isinstance(timeout, (float, int)):
        raise ValueError("Timeout should be a number that define a number of seconds.")
    if poll_interval <= 0 or timeout <= 0 or poll_interval > timeout:
        raise ValueError("Poll interval and timeout should follow this rule (0 < poll interval <= timeout).")


def raise_if_error(response: Response) -> None:
    """Raise an error from a response."""
    if response.is_success:
        return
    error_out = f"HTTP {response.status_code}: {response.reason_phrase}"
    try:
        json_detail = response.json()
    except Exception:
        json_detail = {}

    detail = json_detail.get("detail", None)
    if detail is not None:
        error_out += f" - {detail}"
    error = json_detail.get("error", None)
    if error is not None:
        error_out += f" - {error}"
    raise HTTPException(response.status_code, detail=error_out)


class PixanoInferenceClient(Settings):
    """Pixano Inference Client."""

    url: str

    @field_validator("url", mode="after")
    def _validate_url(cls, v):
        if not is_url(v):
            raise ValueError(f"Invalid URL, got '{v}'.")
        if v.endswith("/"):
            v = v[:-1]
        return v

    @staticmethod
    def connect(url: str) -> "PixanoInferenceClient":
        """Connect to pixano inference.

        Args:
            url: The URL of the pixano inference server.
        """
        settings = Settings.model_validate(requests.get(f"{url}/app/settings/").json())
        client = PixanoInferenceClient(url=url, **settings.model_dump())
        return client

    async def _rest_call(
        self,
        path: str,
        method: Literal["GET", "POST", "PUT", "DELETE"],
        timeout: int = 60,
        **kwargs,
    ) -> Response:
        """Perform a REST call to the pixano inference server."""
        async with httpx.AsyncClient(timeout=timeout) as client:
            match method:
                case "GET":
                    request_fn = client.get
                case "POST":
                    request_fn = client.post
                case "PUT":
                    request_fn = client.put
                case "DELETE":
                    request_fn = client.delete
                case _:
                    raise ValueError(
                        f"Invalid REST call method. Expected one of ['GET', 'POST', 'PUT', 'DELETE'], but got "
                        f"'{method}'."
                    )

            if path.startswith("/"):
                path = path[1:]

            url = f"{self.url}/{path}"
            response = await request_fn(url, **kwargs)
            raise_if_error(response)

            return response

    async def get_settings(self) -> Settings:
        """Get the settings for the pixano inference server."""
        response = await self.get("app/settings/")
        raise_if_error(response)
        return Settings(**response.json())

    async def get(self, path: str, **kwargs: Any) -> Response:
        """Perform a GET request to the pixano inference server.

        Args:
            path: The path of the request.
            kwargs: The keyword arguments to pass to the request or httpx client.
        """
        return await self._rest_call(path=path, method="GET", **kwargs)

    async def post(self, path: str, **kwargs: Any) -> Response:
        """Perform a POST request to the pixano inference server.

        Args:
            path: The path of the request.
            kwargs: The keyword arguments to pass to the request or httpx client.
        """
        return await self._rest_call(path=path, method="POST", **kwargs)

    async def put(self, path: str, **kwargs: Any) -> Response:
        """Perform a PUT request to the pixano inference server.

        Args:
            path: The path of the request.
            kwargs: The keyword arguments to pass to the request or httpx client.
        """
        return await self._rest_call(path=path, method="PUT", **kwargs)

    async def delete(self, path: str, **kwargs: Any) -> Response:
        """Perform a DELETE request to the pixano inference server.

        Args:
            path: The path of the request.
            kwargs: The keyword arguments to pass to the request or httpx client.
        """
        return await self._rest_call(path=path, method="DELETE", **kwargs)

    async def list_models(self) -> list[ModelInfo]:
        """List all models."""
        response = await self.get("app/models/")
        return [ModelInfo.model_construct(**model) for model in response.json()]

    async def instantiate_model(self, provider: str, config: ModelConfig, timeout: int = 60) -> None:
        """Instantiate a model.

        Args:
            provider: The model provider.
            config: The configuration of the model.
            timeout: The timeout to wait for a response. Please note that even if the timeout is reached, the request
                will not be aborted.
        """
        json_content = jsonable_encoder({"provider": provider, "config": config})
        await self.post("providers/instantiate", json=json_content, timeout=timeout)
        return

    async def delete_model(self, model_name: str) -> None:
        """Delete a model.

        Args:
            model_name: The name of the model.
        """
        await self.delete(f"providers/model/{model_name}")

    @overload
    async def inference(
        self,
        route: str,
        request: BaseRequest,
        response_type: type[BaseResponse],
        poll_interval: float,
        timeout: float,
        task_id: str,
        asynchronous: Literal[True],
    ) -> BaseResponse | CeleryTask: ...
    @overload
    async def inference(
        self,
        route: str,
        request: BaseRequest,
        response_type: type[BaseResponse],
        poll_interval: float,
        timeout: float,
        task_id: None,
        asynchronous: Literal[True],
    ) -> CeleryTask: ...
    @overload
    async def inference(
        self,
        route: str,
        request: BaseRequest,
        response_type: type[BaseResponse],
        poll_interval: float,
        timeout: float,
        task_id: str | None,
        asynchronous: Literal[False],
    ) -> BaseResponse: ...
    @overload
    async def inference(
        self,
        route: str,
        request: BaseRequest,
        response_type: type[BaseResponse],
        poll_interval: float,
        timeout: float,
        task_id: str | None,
        asynchronous: bool,
    ) -> BaseResponse | CeleryTask: ...
    async def inference(
        self,
        route: str,
        request: BaseRequest,
        response_type: type[BaseResponse],
        poll_interval: float = 0.1,
        timeout: float = 60.0,
        task_id: str | None = None,
        asynchronous: bool = False,
    ) -> BaseResponse | CeleryTask:
        """Perform a POST request to the pixano inference server.

        Args:
            route: The root to the request.
            request: The request of the model.
            response_type: The type of the response.
            poll_interval: waiting time between subsequent requests to server to retrieve task results for synchronous
                requests.
            timeout: Time to wait for response for synchronous requests. If reached, the request will be aborted.
            task_id: The id of the task to poll for results.
            asynchronous: If True then the function will be called asynchronously and returns a CeleryTask object or
                poll results when task id is provided.

        Returns:
            A response from the pixano inference server.
        """
        _validate_task_id_asynchronous(task_id=task_id, asynchronous=asynchronous)
        _validate_poll_interval_timeout(poll_interval=poll_interval, timeout=timeout)

        if not asynchronous or task_id is None:
            celery_response: Response = await self.post(route, json=request.model_dump())
            celery_task: CeleryTask = CeleryTask.model_construct(**celery_response.json())

        # Asynchronous calls
        if asynchronous and task_id is None:
            return celery_task
        elif asynchronous and task_id is not None:
            has_slash = route.endswith("/")
            task_route = route + f"{'' if has_slash else '/'}{task_id}"
            response: dict[str, Any] = (await self.get(task_route)).json()
            if response["status"] == states.SUCCESS:
                return response_type.model_validate(response)
            return CeleryTask.model_construct(**response)

        # Synchronous calls with polling for result retrieval and deletion of celery tasks after timeout.
        time = 0.0
        has_slash = route.endswith("/")
        task_route = route + f"{'' if has_slash else '/'}{celery_task.id}"
        while time < timeout:
            response = (await self.get(task_route)).json()
            if response["status"] == states.SUCCESS:
                return response_type.model_validate(response)
            elif response["status"] == states.FAILURE:
                raise ValueError("The inference failed. Please check your inputs.")
            time += poll_interval
            await asyncio.sleep(poll_interval)
        await self.delete(task_route)
        raise InferenceTooLongError("The model is either busy or the task takes too long to perform.")

    @overload
    async def text_image_conditional_generation(
        self,
        request: TextImageConditionalGenerationRequest,
        poll_interval: float,
        timeout: float,
        task_id: str,
        asynchronous: Literal[True],
    ) -> TextImageConditionalGenerationResponse | CeleryTask: ...
    @overload
    async def text_image_conditional_generation(
        self,
        request: TextImageConditionalGenerationRequest,
        poll_interval: float,
        timeout: float,
        task_id: None,
        asynchronous: Literal[True],
    ) -> CeleryTask: ...
    @overload
    async def text_image_conditional_generation(
        self,
        request: TextImageConditionalGenerationRequest,
        poll_interval: float,
        timeout: float,
        task_id: str | None,
        asynchronous: Literal[False],
    ) -> TextImageConditionalGenerationResponse | CeleryTask: ...
    @overload
    async def text_image_conditional_generation(
        self,
        request: TextImageConditionalGenerationRequest,
        poll_interval: float,
        timeout: float,
        task_id: str | None,
        asynchronous: bool,
    ) -> TextImageConditionalGenerationResponse | CeleryTask: ...
    async def text_image_conditional_generation(
        self,
        request: TextImageConditionalGenerationRequest,
        poll_interval: float = 0.1,
        timeout: float = 60,
        task_id: str | None = None,
        asynchronous: bool = False,
    ) -> TextImageConditionalGenerationResponse | CeleryTask:
        """Perform an inference to perform text-image conditional generation."""
        return await self.inference(
            route="tasks/multimodal/text-image/conditional_generation/",
            request=request,
            response_type=TextImageConditionalGenerationResponse,
            poll_interval=poll_interval,
            timeout=timeout,
            task_id=task_id,
            asynchronous=asynchronous,
        )

    @overload
    async def image_mask_generation(
        self,
        request: ImageMaskGenerationRequest,
        poll_interval: float,
        timeout: float,
        task_id: str,
        asynchronous: Literal[True],
    ) -> ImageMaskGenerationResponse | CeleryTask: ...
    @overload
    async def image_mask_generation(
        self,
        request: ImageMaskGenerationRequest,
        poll_interval: float,
        timeout: float,
        task_id: None,
        asynchronous: Literal[True],
    ) -> CeleryTask: ...
    @overload
    async def image_mask_generation(
        self,
        request: ImageMaskGenerationRequest,
        poll_interval: float,
        timeout: float,
        task_id: str | None,
        asynchronous: Literal[False],
    ) -> ImageMaskGenerationResponse: ...
    @overload
    async def image_mask_generation(
        self,
        request: ImageMaskGenerationRequest,
        poll_interval: float,
        timeout: float,
        task_id: str | None,
        asynchronous: bool,
    ) -> ImageMaskGenerationResponse | CeleryTask: ...
    async def image_mask_generation(
        self,
        request: ImageMaskGenerationRequest,
        poll_interval: float = 0.1,
        timeout: float = 60,
        task_id: str | None = None,
        asynchronous: bool = False,
    ) -> ImageMaskGenerationResponse | CeleryTask:
        """Perform an inference to perform image mask generation."""
        return await self.inference(
            route="tasks/image/mask_generation/",
            request=request,
            response_type=ImageMaskGenerationResponse,
            poll_interval=poll_interval,
            timeout=timeout,
            task_id=task_id,
            asynchronous=asynchronous,
        )

    @overload
    async def video_mask_generation(
        self,
        request: VideoMaskGenerationRequest,
        poll_interval: float,
        timeout: float,
        task_id: str,
        asynchronous: Literal[True],
    ) -> VideoMaskGenerationResponse | CeleryTask: ...
    @overload
    async def video_mask_generation(
        self,
        request: VideoMaskGenerationRequest,
        poll_interval: float,
        timeout: float,
        task_id: None,
        asynchronous: Literal[True],
    ) -> CeleryTask: ...
    @overload
    async def video_mask_generation(
        self,
        request: VideoMaskGenerationRequest,
        poll_interval: float,
        timeout: float,
        task_id: str | None,
        asynchronous: Literal[False],
    ) -> VideoMaskGenerationResponse: ...
    @overload
    async def video_mask_generation(
        self,
        request: VideoMaskGenerationRequest,
        poll_interval: float,
        timeout: float,
        task_id: str | None,
        asynchronous: bool,
    ) -> VideoMaskGenerationResponse | CeleryTask: ...
    async def video_mask_generation(
        self,
        request: VideoMaskGenerationRequest,
        poll_interval: float = 0.1,
        timeout: float = 60,
        task_id: str | None = None,
        asynchronous: bool = False,
    ) -> VideoMaskGenerationResponse | CeleryTask:
        """Perform an inference to perform video mask generation."""
        return await self.inference(
            route="tasks/video/mask_generation/",
            request=request,
            response_type=VideoMaskGenerationResponse,
            poll_interval=poll_interval,
            timeout=timeout,
            task_id=task_id,
            asynchronous=asynchronous,
        )

    @overload
    async def image_zero_shot_detection(
        self,
        request: ImageZeroShotDetectionRequest,
        poll_interval: float,
        timeout: float,
        task_id: str,
        asynchronous: Literal[True],
    ) -> ImageZeroShotDetectionResponse | CeleryTask: ...
    @overload
    async def image_zero_shot_detection(
        self,
        request: ImageZeroShotDetectionRequest,
        poll_interval: float,
        timeout: float,
        task_id: None,
        asynchronous: Literal[True],
    ) -> CeleryTask: ...
    @overload
    async def image_zero_shot_detection(
        self,
        request: ImageZeroShotDetectionRequest,
        poll_interval: float,
        timeout: float,
        task_id: str | None,
        asynchronous: Literal[False],
    ) -> ImageZeroShotDetectionResponse: ...
    @overload
    async def image_zero_shot_detection(
        self,
        request: ImageZeroShotDetectionRequest,
        poll_interval: float,
        timeout: float,
        task_id: str | None,
        asynchronous: bool,
    ) -> ImageZeroShotDetectionResponse | CeleryTask: ...
    async def image_zero_shot_detection(
        self,
        request: ImageZeroShotDetectionRequest,
        poll_interval: float = 0.1,
        timeout: float = 60,
        task_id: str | None = None,
        asynchronous: bool = False,
    ) -> ImageZeroShotDetectionResponse | CeleryTask:
        """Perform an inference to perform video mask generation."""
        return await self.inference(
            route="tasks/image/zero_shot_detection/",
            request=request,
            response_type=ImageZeroShotDetectionResponse,
            poll_interval=poll_interval,
            timeout=timeout,
            task_id=task_id,
            asynchronous=asynchronous,
        )



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""This module contains functions for reading data from Vector databases."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal

import numpy as np
import polars as pl

from pixano_inference.utils.package import (
    assert_lance_installed,
    assert_torch_installed,
    is_lance_installed,
    is_torch_installed,
)


if is_lance_installed():
    from lance import LanceDataset

if is_torch_installed():
    import torch


if TYPE_CHECKING:
    from torch import Tensor


def read_lance_vector(
    path: Path | str,
    column: str,
    indice: int | None = None,
    where: str | None = None,
    shape: list[int] | None = None,
    return_type: Literal["tensor", "numpy"] = "tensor",
) -> "Tensor" | np.ndarray:
    """Reads a vector from a Lance dataset.

    Args:
        path: The path to the Lance dataset.
        column: The vector column to read.
        indice: The index of the row to read.
        where: The filter to apply. If specified, only one row should be returned.
        shape: The shape of the vector.
        return_type: The type of the return value. Either 'tensor' or 'numpy'.

    Returns:
        The value of the column.
    """
    assert_lance_installed()
    if return_type not in ["tensor", "numpy"]:
        raise ValueError("return_type must be either 'tensor' or 'numpy'.")
    elif return_type == "tensor":
        assert_torch_installed()
    if indice is not None and where is not None:
        raise ValueError("Only one of 'index' and 'where' can be specified.")
    elif indice is None and where is None:
        raise ValueError("One of 'index' and 'where' must be specified.")
    elif indice is not None and not isinstance(indice, int):
        raise ValueError("index must be an integer.")
    elif where is not None and not isinstance(where, str):
        raise ValueError("where must be a string.")
    elif not isinstance(path, (Path, str)):
        raise ValueError("lance_path must be a Path or a string.")
    elif shape is not None and not isinstance(shape, list) and not all(isinstance(i, int) for i in shape):
        raise ValueError("shape must be a list of integers.")

    if isinstance(path, str):
        path = Path(path)

    lance_dataset = LanceDataset(path)
    if indice is not None:
        pa_table = lance_dataset.take(indices=[indice], columns=[column])
    else:
        pa_table = lance_dataset.scanner(columns=[column], filter=where).to_table()
    polar_df = pl.DataFrame(pa_table)
    if polar_df.shape[0] != 1:
        raise ValueError(f"Expected one row to be returned but found {polar_df.shape[0]} rows.")
    vector: np.ndarray = polar_df[column].to_numpy()
    vector = vector.reshape(shape)
    if return_type == "tensor":
        return torch.from_numpy(vector)
    return vector



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

# ruff: noqa: F401
# ruff: noqa: D104

from .read_vector_databases import read_lance_vector



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Pydantic models for model configuration."""

from pathlib import Path
from typing import Any

from pydantic import BaseModel, field_validator

from pixano_inference.tasks.utils import is_task


class ModelInfo(BaseModel):
    """Model Information.

    Attributes:
        name: Name of the model.
        task: Task of the model.
    """

    name: str
    task: str


class ModelConfig(BaseModel):
    """Model configuration for instantiation.

    Attributes:
        name: Name of the model.
        task: Task of the model.
        path: Path to the model dump.
        config: Configuration of the model.
        processor_config: Configuration of the processor.
    """

    name: str
    task: str
    path: Path | str | None = None
    config: dict[str, Any] = {}
    processor_config: dict[str, Any] = {}

    @field_validator("task")
    def _check_task(cls, value):
        if not is_task(value):
            raise ValueError(f"Invalid task '{value}'")
        return value



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""API routes for Image tasks."""

from typing import Annotated

from fastapi import APIRouter, Depends

from pixano_inference.pydantic.base import CeleryTask
from pixano_inference.pydantic.tasks.video.mask_generation import (
    VideoMaskGenerationRequest,
)
from pixano_inference.routers.utils import execute_task_request
from pixano_inference.settings import Settings, get_pixano_inference_settings
from pixano_inference.tasks.video import VideoTask


router = APIRouter(prefix="/video", tags=["Video"])


@router.post("/mask_generation", response_model=CeleryTask)
async def mask_generation(
    request: VideoMaskGenerationRequest,
    settings: Annotated[Settings, Depends(get_pixano_inference_settings)],
) -> CeleryTask:
    """Generate mask from a video and optionnaly points and bboxes.

    Args:
        request: Request for mask generation.
        settings: Settings of the app.

    Returns:
        Response for mask generation.
    """
    return await execute_task_request(request=request, task=VideoTask.MASK_GENERATION, settings=settings)



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Utility functions for working with Python packages."""

import importlib.util


def is_package_installed(package_name: str) -> bool:
    """Check if a Python package is installed.

    Args:
        package_name: The name of the package to check.

    Returns:
        True if the package is installed, False otherwise
    """
    package_spec = importlib.util.find_spec(package_name)
    return package_spec is not None


def is_lance_installed() -> bool:
    """Check if the lance package is installed.

    Returns:
        True if the lance package is installed, False otherwise
    """
    return is_package_installed("lance")


def is_sam2_installed() -> bool:
    """Check if the sam2 package is installed.

    Returns:
        True if the sam2 package is installed, False otherwise
    """
    return is_package_installed("sam2")


def is_torch_installed() -> bool:
    """Check if the torch package is installed.

    Returns:
        True if the torch package is installed, False otherwise
    """
    return is_package_installed("torch")


def is_transformers_installed() -> bool:
    """Check if the transformers package is installed.

    Returns:
        True if the transformers package is installed, False otherwise
    """
    return is_package_installed("transformers")


def is_vllm_installed() -> bool:
    """Check if the vllm package is installed.

    Returns:
        True if the vLLM package is installed, False otherwise
    """
    return is_package_installed("vllm")


def assert_package_installed(package_name: str, error_message: str | None = None) -> None:
    """Assert that a Python package is installed.

    Args:
        package_name: The name of the package to check.
        error_message: The error message to raise if the package is not installed.

    Raises:
        ImportError: If the package is not installed
    """
    if not is_package_installed(package_name):
        message = error_message or f"Package '{package_name}' is not installed."
        raise ImportError(message)


def assert_lance_installed() -> None:
    """Assert that the lance package is installed."""
    assert_package_installed(
        "lance", "lance is not installed. Please install it using 'pip install pixano-inference[data]'."
    )


def assert_sam2_installed() -> None:
    """Assert that the sam2 package is installed."""
    assert_package_installed(
        "sam2",
        "sam2 is not installed. Please install it using 'pip install git+https://github.com/facebookresearch/sam2.git'.",
    )


def assert_transformers_installed() -> None:
    """Assert that the transformers package is installed."""
    assert_package_installed(
        "transformers",
        "transformers is not installed. Please install it using 'pip install pixano-inference[transformers]'.",
    )


def assert_torch_installed() -> None:
    """Assert that the torch package is installed."""
    assert_package_installed(
        "torch", "torch is not installed. Please install it using 'pip install pixano-inference[torch]'."
    )


def assert_vllm_installed() -> None:
    """Assert that the vllm package is installed."""
    assert_package_installed(
        "vllm", "vLLM is not installed. Please install it using 'pip install pixano-inference[vllm]'."
    )



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Task utilities."""

from .image import ImageTask
from .multimodal import MultimodalImageNLPTask
from .nlp import NLPTask
from .task import Task
from .video import VideoTask


STR_IMAGE_TASKS = {task.value for task in ImageTask}
STR_NLP_TASKS = {task.value for task in NLPTask}
STR_MULTIMODAL_TASKS = {task.value for task in MultimodalImageNLPTask}
STR_VIDEO_TASKS = {task.value for task in VideoTask}


def get_tasks() -> set[str]:
    """Get all tasks."""
    return {task for set in [STR_IMAGE_TASKS, STR_NLP_TASKS, STR_MULTIMODAL_TASKS, STR_VIDEO_TASKS] for task in set}


def is_task(task: str) -> bool:
    """Check if a task is valid."""
    return task in get_tasks()


def str_to_task(task: str) -> Task:
    """Convert a task string to its task."""
    if task in STR_IMAGE_TASKS:
        return ImageTask(task)
    elif task in STR_NLP_TASKS:
        return NLPTask(task)
    elif task in STR_MULTIMODAL_TASKS:
        return MultimodalImageNLPTask(task)
    elif task in STR_VIDEO_TASKS:
        return VideoTask(task)
    raise ValueError(f"Invalid task '{task}'")



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Provider for the SAM2 model."""

from pathlib import Path
from typing import Any

from pixano_inference.models.base import BaseInferenceModel
from pixano_inference.models.sam2 import Sam2Model
from pixano_inference.providers.registry import register_provider
from pixano_inference.pydantic.tasks.image.mask_generation import ImageMaskGenerationOutput, ImageMaskGenerationRequest
from pixano_inference.pydantic.tasks.video.mask_generation import (
    VideoMaskGenerationRequest,
    VideoMaskGenerationResponse,
)
from pixano_inference.tasks.image import ImageTask
from pixano_inference.tasks.task import Task
from pixano_inference.tasks.utils import str_to_task
from pixano_inference.tasks.video import VideoTask
from pixano_inference.utils.image import convert_string_to_image
from pixano_inference.utils.package import assert_sam2_installed, is_torch_installed
from pixano_inference.utils.vector import vector_to_tensor

from .base import ModelProvider


if is_torch_installed():
    import torch


@register_provider("sam2")
class Sam2Provider(ModelProvider):
    """Provider for the SAM2 model."""

    def __init__(self, **kwargs):
        """Initialize the SAM2 provider."""
        assert_sam2_installed()
        super().__init__(**kwargs)

    def load_model(
        self,
        name: str,
        task: Task | str,
        device: "torch.device",
        path: Path | str | None = None,
        processor_config: dict = {},
        config: dict = {},
    ) -> Sam2Model:
        """Load the model.

        Args:
            name: Name of the model.
            task: Task of the model.
            device: Device to use for the model.
            path: Path to the model.
            processor_config: Processor configuration.
            config: Configuration for the model.

        Returns:
            The loaded model.
        """
        from sam2.build_sam import build_sam2, build_sam2_hf, build_sam2_video_predictor, build_sam2_video_predictor_hf
        from sam2.sam2_image_predictor import SAM2ImagePredictor

        task = str_to_task(task) if isinstance(task, str) else task
        if task == ImageTask.MASK_GENERATION:
            if path is not None and Path(path).exists():
                model = build_sam2(ckpt_path=path, mode="eval", device=device, **config)
            else:
                model = build_sam2_hf(model_id=path, mode="eval", device=device, **config)
            model = torch.compile(model)
            predictor = SAM2ImagePredictor(model)
        elif task == VideoTask.MASK_GENERATION:
            if path is not None and Path(path).exists():
                predictor = build_sam2_video_predictor(ckpt_path=path, mode="eval", device=device, **config)
            else:
                predictor = build_sam2_video_predictor_hf(model_id=path, mode="eval", device=device, **config)
            predictor = torch.compile(predictor)
        else:
            raise ValueError(f"Invalid task '{task}' for the SAM2 provider.")

        our_model = Sam2Model(
            name=name,
            provider="sam2",
            predictor=predictor,
            torch_dtype=config.get("torch_dtype", "bfloat16"),
            config=config,
        )

        return our_model

    def image_mask_generation(
        self,
        request: ImageMaskGenerationRequest,
        model: Sam2Model,  # type: ignore[override]
        *args: Any,
        **kwargs: Any,
    ) -> ImageMaskGenerationOutput:
        """Generate a mask from the image.

        Args:
            request: Request for the generation.
            model: Model to use for the generation.
            args: Additional arguments.
            kwargs: Additional keyword arguments.

        Returns:
            Output of the generation
        """
        request_input = request.to_input()
        image = convert_string_to_image(request_input.image)

        if request_input.image_embedding is not None and request_input.high_resolution_features is not None:
            image_embedding = vector_to_tensor(request_input.image_embedding)
            high_resolution_features = [vector_to_tensor(v) for v in request_input.high_resolution_features]
            model.set_image_embeddings(image, image_embedding, high_resolution_features)
        elif request_input.image_embedding is not None or request_input.high_resolution_features is not None:
            raise ValueError("Both image_embedding and high_resolution_features must be provided.")

        model_input = request_input.model_dump(exclude=["image", "image_embedding", "high_resolution_features"])
        model_input["image"] = image
        output = model.image_mask_generation(**model_input)
        model.predictor.reset_predictor()
        return output

    def video_mask_generation(
        self, request: VideoMaskGenerationRequest, model: BaseInferenceModel, *args: Any, **kwargs: Any
    ) -> VideoMaskGenerationResponse:
        """Generate masks from the video.

        Args:
            request: Request for the generation.
            model: Model to use for the generation.
            args: Additional arguments.
            kwargs: Additional keyword arguments.

        Returns:
            Response of the generation.
        """
        request_input = request.to_input()
        output = model.video_mask_generation(**request_input)
        return output



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""API routes for Image tasks."""

from fastapi import APIRouter

from .image import router as image_router
from .multimodal import router as multimodal_router
from .nlp import router as nlp_router
from .video import router as video_router


router = APIRouter(prefix="/tasks", tags=["Tasks"])
router.include_router(image_router)
router.include_router(multimodal_router)
router.include_router(nlp_router)
router.include_router(video_router)



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Image tasks."""

from .task import Task


class ImageTask(Task):
    """Image tasks."""

    CLASSIFICATION = "image_classification"
    DEPTH_ESTIMATION = "depth_estimation"
    INSTANCE_SEGMENTATION = "instance_segmentation"
    FEATURE_EXTRACTION = "image_feature_extraction"
    KEYPOINT_DETECTION = "keypoint_detection"
    MASK_GENERATION = "image_mask_generation"
    OBJECT_DETECTION = "object_detection"
    SEMANTIC_SEGMENTATION = "semantic_segmentation"
    UNIVERSAL_SEGMENTATION = "universal_segmentation"
    ZERO_SHOT_CLASSIFICATION = "image_zero_shot_classification"
    ZERO_SHOT_DETECTION = "image_zero_shot_detection"



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

import pytest

from pixano_inference.tasks import ImageTask, MultimodalImageNLPTask, NLPTask, VideoTask
from pixano_inference.tasks.utils import get_tasks, is_task, str_to_task


def test_is_task():
    assert is_task("image_classification")
    assert not is_task("not_a_task")


def test_get_tasks():
    tasks = get_tasks()
    assert len(tasks) == len(ImageTask) + len(NLPTask) + len(MultimodalImageNLPTask) + len(VideoTask)
    assert tasks == {task.value for task in ImageTask} | {task.value for task in NLPTask} | {
        task.value for task in MultimodalImageNLPTask
    } | {task.value for task in VideoTask}


def test_str_to_task():
    assert str_to_task("image_classification") == ImageTask.CLASSIFICATION
    assert str_to_task("token_classification") == NLPTask.TOKEN_CLASSIFICATION
    assert str_to_task("image_captioning") == MultimodalImageNLPTask.CAPTIONING
    assert str_to_task("video_mask_generation") == VideoTask.MASK_GENERATION

    with pytest.raises(ValueError, match="Invalid task 'not_a_task'"):
        str_to_task("not_a_task")



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Utility functions for vector operations."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pixano_inference.pydantic.data.vector_database import LanceVector
from pixano_inference.pydantic.nd_array import NDArrayFloat
from pixano_inference.utils.package import assert_torch_installed, is_torch_installed


if is_torch_installed():
    from torch import Tensor

if TYPE_CHECKING:
    from torch import Tensor


def vector_to_tensor(vector: NDArrayFloat | LanceVector | "Tensor" | None) -> "Tensor":
    """Convert a vector to a tensor.

    Args:
        vector: Vector to convert.
    """
    assert_torch_installed()
    if not isinstance(vector, (NDArrayFloat, LanceVector, Tensor)) and vector is not None:
        raise ValueError(f"Unsupported vector type: {type(vector)}")

    if isinstance(vector, LanceVector):
        return vector.read_vector()
    elif isinstance(vector, NDArrayFloat):
        return vector.to_torch()
    return vector



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

# ruff: noqa: F401
# ruff: noqa: D104

from .image import ImageTask
from .multimodal import MultimodalImageNLPTask
from .nlp import NLPTask
from .task import Task
from .utils import STR_IMAGE_TASKS, STR_MULTIMODAL_TASKS, STR_NLP_TASKS, get_tasks, is_task, str_to_task
from .video import VideoTask



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

# ruff: noqa: F401
# ruff: noqa: D104

from .base import BaseProvider
from .registry import register_provider
from .sam2 import Sam2Provider
from .transformers import TransformersProvider
from .utils import get_provider, get_providers, instantiate_provider, is_provider
from .vllm import VLLMProvider



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Registry for providers."""

from pixano_inference.providers.base import BaseProvider


PROVIDERS_REGISTRY: dict[str, type[BaseProvider]] = {}


def register_provider(provider: str):
    """Return a decorator to register a provider in the registry.

    Args:
        provider: Name of the provider.

    Returns:
        The decorator.
    """

    def decorator(cls):
        """Register the provider in the registry.

        Args:
            cls: Class to register.

        Returns:
            The class.
        """
        if provider in PROVIDERS_REGISTRY:
            raise ValueError(f"Provider {provider} already registered.")
        PROVIDERS_REGISTRY[provider] = cls
        return cls

    return decorator



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Image utilities."""

import base64
import re
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
import requests  # type: ignore[import-untyped]
from PIL import Image

from .package import assert_torch_installed, is_torch_installed
from .url import is_url


if is_torch_installed():
    import torch

if TYPE_CHECKING:
    from torch import Tensor

regex_base64 = r"^(data:image/[a-zA-Z]+;base64,)"


def is_base64_image(image: str) -> bool:
    """Check if a string is a base64 image.

    The expected format is "data:image/{image_format};base64,{base64}".
    """
    match = re.match(regex_base64, image)
    return match is not None


def extract_image_from_base64(image: str):
    """Extract from a base64 image the actual base64 part."""
    match = re.match(regex_base64, image)
    if match is None:
        return ""
    return image[len(match.group(1)) :]


def convert_string_to_image(str_image: str | Path) -> Image.Image:
    """Convert a string or path to an image.

    Args:
        str_image: Image as a string or path.

    Returns:
        Image.
    """
    if isinstance(str_image, str):
        if is_url(str_image):
            image_pil = Image.open(requests.get(str_image, stream=True).raw)
        else:
            if is_base64_image(str_image):
                image_bytes = base64.b64decode(extract_image_from_base64(str_image))
                image_pil = Image.open(BytesIO(image_bytes))
            elif Path(str_image).exists():
                image_pil = Image.open(str_image)
            else:
                raise ValueError("The image is not a valid path, URL or base64 string.")
    elif isinstance(str_image, Path):
        image_pil = Image.open(str_image)
    else:
        raise ValueError("The image is not a valid path, URL or base64 string.")
    image_converted = image_pil.convert("RGB")
    return image_converted


def encode_mask_to_rle(mask: "Tensor") -> dict[str, list[int]]:
    """Encode a binary mask using RLE.

    Args:
        mask: A binary mask of shape (height, width).

    Returns:
        RLE encoded mask as a dictionary.
    """
    assert_torch_installed()
    rle = {"counts": [], "size": list(mask.shape)}
    mask = mask.permute(1, 0).flatten()
    diff_arr = torch.diff(mask)
    nonzero_indices = torch.where(diff_arr != 0)[0] + 1
    lengths = torch.diff(torch.concatenate((torch.tensor([0]), nonzero_indices, torch.tensor([len(mask)]))))

    # note that the odd counts are always the numbers of zeros
    if mask[0] == 1:
        lengths = torch.concatenate(([0], lengths))

    rle["counts"] = lengths.tolist()

    return rle


def compress_rle(rle: dict[str, Any]) -> dict[str, Any | str]:
    """Compress an RLE encoded mask.

    Args:
        rle: RLE encoded mask as a dictionary.

    Returns:
        Compressed RLE encoded mask as a string.
    """
    counts = np.array(rle["counts"], dtype=np.uint32).tobytes()
    rle["counts"] = base64.b64encode(counts).decode("utf-8")
    return rle


def decode_rle_to_mask(rle: dict) -> np.ndarray:
    """Decode an RLE encoded mask.

    Args:
        rle: RLE encoded mask as a dictionary.

    Returns:
        Decoded binary mask of shape (height, width).
    """
    height, width = rle["size"]
    mask = np.empty(height * width, dtype=bool)
    idx = 0
    parity = False
    for count in rle["counts"]:
        mask[idx : idx + count] = parity
        idx += count
        parity = not parity
    mask = mask.reshape(width, height)
    return mask.transpose()  # Reshape to original shape


def decompress_rle(rle: dict[str, Any]) -> dict[str, Any]:
    """Decompress a compressed RLE encoded mask.

    Args:
        rle: Compressed RLE encoded mask as a string.

    Returns:
        Decompressed RLE encoded mask as a dictionary.
    """
    rle["counts"] = np.frombuffer(base64.b64decode(rle["counts"]), dtype=np.uint32).tolist()
    return rle



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================


import re

import pytest
import responses
from fastapi import HTTPException
from pytest_httpx import HTTPXMock

from pixano_inference.client import PixanoInferenceClient
from pixano_inference.pydantic import ModelInfo
from pixano_inference.settings import Settings


URL = "http://localhost:8081"


class TestPixanoInferenceClient:
    def test_init_client(self):
        client = PixanoInferenceClient(url=URL)
        assert client.url == URL

        client = PixanoInferenceClient(url=f"{URL}/")
        assert client.url == URL

        with pytest.raises(ValueError, match="Invalid URL, got 'wrongurl'."):
            client = PixanoInferenceClient(url="wrongurl")

    @responses.activate
    def test_connect(self):
        settings = Settings(num_cpus=1, num_gpus=0)
        response = responses.Response(method="GET", url=f"{URL}/app/settings/", json=settings.model_dump(), status=200)
        responses.add(response)
        expected_output = settings.model_dump()
        expected_output.update({"url": URL})

        client = PixanoInferenceClient.connect(URL)
        assert client.model_dump() == expected_output

    @pytest.mark.asyncio
    @pytest.mark.parametrize("method", ["GET", "POST", "PUT", "DELETE"])
    async def test_valid_rest_call(
        self, httpx_mock: HTTPXMock, simple_pixano_inference_client: PixanoInferenceClient, method: str
    ):
        data = {"image_url": "https://example.com/image.jpg"}
        httpx_mock.add_response(json=data)
        output_response = await simple_pixano_inference_client._rest_call("data/", method)
        assert output_response.status_code == 200
        assert output_response.json() == data

    @pytest.mark.asyncio
    async def test_invalid_rest_call(
        self, httpx_mock: HTTPXMock, simple_pixano_inference_client: PixanoInferenceClient
    ):
        with pytest.raises(
            ValueError,
            match=re.escape(
                r"Invalid REST call method. Expected one of ['GET', 'POST', 'PUT', 'DELETE'], "
                r"but got 'WRONG_METHOD'."
            ),
        ):
            await simple_pixano_inference_client._rest_call("data/", "WRONG_METHOD")

        httpx_mock.add_response(json={"error": "Invalid data."}, status_code=400)

        with pytest.raises(HTTPException) as exc_info:
            await simple_pixano_inference_client._rest_call("data/", "POST")
            assert exc_info.value.status_code == 400
            assert exc_info.value.response.json() == {
                "error": "Invalid data.",
            }

    @pytest.mark.asyncio
    async def test_get(self, httpx_mock: HTTPXMock, simple_pixano_inference_client: PixanoInferenceClient):
        data = {"image_url": "https://example.com/image.jpg"}
        httpx_mock.add_response(json=data)
        assert (await simple_pixano_inference_client.get("data/")).json() == data

    @pytest.mark.asyncio
    async def test_post(self, httpx_mock: HTTPXMock, simple_pixano_inference_client: PixanoInferenceClient):
        data = {"image_url": "https://example.com/image.jpg"}
        httpx_mock.add_response(json=data)
        assert (await simple_pixano_inference_client.post("data/", json=data)).json() == data

    @pytest.mark.asyncio
    async def test_put(self, httpx_mock: HTTPXMock, simple_pixano_inference_client: PixanoInferenceClient):
        data = {"image_url": "https://example.com/image.jpg"}
        httpx_mock.add_response(json=data)
        assert (await simple_pixano_inference_client.put("data/", json=data)).json() == data

    @pytest.mark.asyncio
    async def test_delete(self, httpx_mock: HTTPXMock, simple_pixano_inference_client: PixanoInferenceClient):
        httpx_mock.add_response()
        assert (await simple_pixano_inference_client.delete("data/")).status_code == 200

    @pytest.mark.asyncio
    async def test_list_models(self, httpx_mock: HTTPXMock, simple_pixano_inference_client: PixanoInferenceClient):
        models_info = [
            ModelInfo(**{"name": "sam", "provider": "transformers", "task": "image_mask_generation"}),
            ModelInfo(**{"name": "sam2", "provider": "sam2", "task": "video_mask_generation"}),
        ]
        httpx_mock.add_response(json=[m.model_dump() for m in models_info])

        assert (await simple_pixano_inference_client.list_models()) == models_info

    @pytest.mark.asyncio
    async def test_delete_model(self, httpx_mock: HTTPXMock, simple_pixano_inference_client: PixanoInferenceClient):
        httpx_mock.add_response()
        assert await simple_pixano_inference_client.delete_model("model_name") is None



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

# ruff: noqa: F401
# ruff: noqa: D104

from .mask_generation import (
    ImageMaskGenerationInput,
    ImageMaskGenerationOutput,
    ImageMaskGenerationRequest,
    ImageMaskGenerationResponse,
)
from .utils import CompressedRLE
from .zero_shot_detection import (
    ImageZeroShotDetectionInput,
    ImageZeroShotDetectionOutput,
    ImageZeroShotDetectionRequest,
    ImageZeroShotDetectionResponse,
)



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""API routes for NLP tasks."""

from fastapi import APIRouter


router = APIRouter(prefix="/nlp", tags=["NLP"])



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Pydantic models for mask generation task."""

from pathlib import Path

from pydantic import BaseModel, ConfigDict, field_validator

from pixano_inference.pydantic.base import BaseRequest, BaseResponse
from pixano_inference.pydantic.tasks.image.utils import CompressedRLE


class VideoMaskGenerationInput(BaseModel):
    """Input for mask generation.

    Attributes:
        frames: Frames for mask generation.
        points: Points for the mask generation. The first fimension is the number of objects the second
            the number of points for each object and the third the coordinates of the points.
        labels: Labels for the mask generation. The first fimension is the number of objects, the second
            the number of labels for each object.
        boxes: Boxes for the mask generation. The first fimension is the number of objects, the second
            the coordinates of the boxes.
        objects_ids: IDs of the objects to generate masks for.
        frame_indexes: Indexes of the frames where the objects are located.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    video: str | Path
    points: list[list[list[int]]] | None = None
    labels: list[list[int]] | None = None
    boxes: list[list[int]] | None = None
    objects_ids: list[int]
    frame_indexes: list[int]

    @field_validator("points")
    @classmethod
    def _check_points(cls, v: list[list[list[int]]] | None) -> list[list[list[int]]] | None:
        if v is not None:
            for list_ in v:
                for point in list_:
                    if len(point) != 2:
                        raise ValueError("Each point should have 2 coordinates.")
        return v

    @field_validator("boxes")
    @classmethod
    def _check_boxes(cls, v: list[list[int]] | None) -> list[list[int]] | None:
        if v is not None:
            for box in v:
                if len(box) != 4:
                    raise ValueError("Each box should have 4 coordinates.")
        return v

    @field_validator("objects_ids")
    @classmethod
    def _check_objects_ids(cls, v: list[int]) -> list[int]:
        if len(v) == 0:
            raise ValueError("At least one object ID should be provided.")
        elif len(v) != len(set(v)):
            raise ValueError("Object IDs should be unique.")
        return v


class VideoMaskGenerationRequest(BaseRequest, VideoMaskGenerationInput):
    """Request for mask generation."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def to_input(self) -> VideoMaskGenerationInput:
        """Convert the request to the input."""
        return self.to_base_model(VideoMaskGenerationInput)


class VideoMaskGenerationOutput(BaseModel):
    """Output for mask generation.

    Attributes:
        objects_ids: IDs of the objects.
        frame_indexes: Indexes of the frames where the objects are located.
        masks: Masks for the objects.
    """

    objects_ids: list[int]
    frame_indexes: list[int]
    masks: list[CompressedRLE]


class VideoMaskGenerationResponse(BaseResponse):
    """Response for mask generation.

    Attributes:
        data: Output of the generation.
    """

    data: VideoMaskGenerationOutput



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Pydantic models for image mask generation task."""

from pathlib import Path

from pydantic import BaseModel, ConfigDict, field_validator

from pixano_inference.pydantic.base import BaseRequest, BaseResponse
from pixano_inference.pydantic.data.vector_database import LanceVector
from pixano_inference.pydantic.nd_array import NDArrayFloat

from .utils import CompressedRLE


class ImageMaskGenerationInput(BaseModel):
    """Input for image mask generation.

    Attributes:
        image: Image for image mask generation.
        image_embedding: Image embedding for the image mask generation.
        high_resolution_features: High resolution features for the image mask generation.
        points: Points for the image mask generation. The first fimension is the number of prompts the second
            the number of points per mask and the third the coordinates of the points.
        labels: Labels for the image mask generation. The first fimension is the number of prompts, the second
            the number of labels per mask.
        boxes: Boxes for the image mask generation. The first fimension is the number of prompts, the second
            the coordinates of the boxes.
        num_multimask_outputs: Number of masks to generate per prediction.
        multimask_output: Whether to generate multiple masks per prediction.
        return_image_embedding: Whether to return the image embeddings.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    image: str | Path
    image_embedding: NDArrayFloat | LanceVector | None = None
    high_resolution_features: list[NDArrayFloat] | list[LanceVector] | None = None
    points: list[list[list[int]]] | None = None
    labels: list[list[int]] | None = None
    boxes: list[list[int]] | None = None
    num_multimask_outputs: int = 3
    multimask_output: bool = True
    return_image_embedding: bool = False

    @field_validator("points")
    @classmethod
    def _check_points(cls, v: list[list[list[int]]] | None) -> list[list[list[int]]] | None:
        if v is not None:
            for list_ in v:
                for point in list_:
                    if len(point) != 2:
                        raise ValueError("Each point should have 2 coordinates.")
        return v

    @field_validator("boxes")
    @classmethod
    def _check_boxes(cls, v: list[list[int]] | None) -> list[list[int]] | None:
        if v is not None:
            for box in v:
                if len(box) != 4:
                    raise ValueError("Each box should have 4 coordinates.")
        return v


class ImageMaskGenerationRequest(BaseRequest, ImageMaskGenerationInput):
    """Request for image mask generation."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def to_input(self) -> ImageMaskGenerationInput:
        """Convert the request to the input."""
        return self.to_base_model(ImageMaskGenerationInput)


class ImageMaskGenerationOutput(BaseModel):
    """Output for image mask generation.

    Attributes:
        masks: Generated masks. The first dimension is the number of predictions, the second the number of masks per
            prediction.
        scores: Scores of the masks. The first dimension is the number of predictions, the second the number of masks
            per prediction.
        image_embedding: Image embeddings.
        high_resolution_features: High resolution features.
    """

    masks: list[list[CompressedRLE]]
    scores: NDArrayFloat
    image_embedding: NDArrayFloat | None = None
    high_resolution_features: list[NDArrayFloat] | None = None


class ImageMaskGenerationResponse(BaseResponse):
    """Response for image mask generation.

    Attributes:
        data: Output of the generation.
    """

    data: ImageMaskGenerationOutput



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

# ruff: noqa: F401
# ruff: noqa: D104

from .image import convert_string_to_image, decode_rle_to_mask, encode_mask_to_rle
from .url import is_url
from .vector import vector_to_tensor



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Inference models for Transformers."""

from __future__ import annotations

import gc
from pathlib import Path
from typing import TYPE_CHECKING, Any

from pixano_inference.models.base import BaseInferenceModel
from pixano_inference.pydantic import (
    CompressedRLE,
    ImageMaskGenerationOutput,
    ImageZeroShotDetectionOutput,
    NDArrayFloat,
    TextImageConditionalGenerationOutput,
    UsageConditionalGeneration,
)
from pixano_inference.utils.image import encode_mask_to_rle
from pixano_inference.utils.package import (
    assert_transformers_installed,
    is_torch_installed,
    is_transformers_installed,
)


if is_torch_installed():
    import torch

if is_transformers_installed():
    import transformers
    from transformers import GenerationConfig


if TYPE_CHECKING:
    from PIL.Image import Image
    from torch import Tensor
    from transformers import PreTrainedModel, ProcessorMixin


class TransformerModel(BaseInferenceModel):
    """Inference model for transformers."""

    def __init__(self, name: str, path: Path | str, processor: "ProcessorMixin", model: "PreTrainedModel"):
        """Initialize the model.

        Args:
            name: Name of the model.
            path: Path to the model or its Hugging Face hub's identifier.
            processor: Processor for the model.
            model: Model for the inference.
        """
        assert_transformers_installed()

        super().__init__(name, provider="transformers")
        self.processor = processor
        self.path = path
        self.model = model.eval()

    @property
    def metadata(self) -> dict[str, Any]:
        """Return the metadata of the model."""
        return {
            "name": self.name,
            "path": self.path,
            "provider": self.provider,
            "provider_version": transformers.__version__,
            "processor_config": self.processor.to_dict(),
            "model_config": self.model.config.to_dict(),
        }

    def _fill_generation_config(self, generation_config: "GenerationConfig", **kwargs) -> "GenerationConfig":
        generation_config = generation_config.to_dict()
        for kwarg in kwargs:
            if kwarg in generation_config:
                generation_config[kwarg] = kwargs[kwarg]
        return GenerationConfig.from_dict(generation_config)

    def delete(self):
        """Delete the model."""
        del self.model
        del self.processor
        gc.collect()
        torch.cuda.empty_cache()

    def image_mask_generation(
        self,
        image: "Tensor" | Image,
        image_embedding: "Tensor" | None = None,
        points: list[list[list[int]]] | None = None,
        labels: list[list[int]] | None = None,
        boxes: list[list[int]] | None = None,
        num_multimask_outputs: int = 3,
        multimask_output: bool = True,
        return_image_embedding: bool = False,
        **kwargs: Any,
    ) -> ImageMaskGenerationOutput:
        """Generate a mask from the image.

        Args:
            image: Image for the generation.
            image_embedding: Image embeddings for the generation.
            points: Points for the mask generation. The first fimension is the number of prompts, the
                second the number of points per mask and the third the coordinates of the points.
            labels: Labels for the mask generation. The first fimension is the number of prompts, the second
                the number of labels per mask.
            boxes: Boxes for the mask generation. The first fimension is the number of prompts, the second
                the coordinates of the boxes.
            num_multimask_outputs: Number of masks to generate per prediction.
            multimask_output: Whether to generate multiple masks per prediction.
            return_image_embedding: Whether to return the image embedding.
            kwargs: Additional keyword arguments.
        """
        with torch.inference_mode():
            inputs = self.processor(
                image,
                input_points=[points] if points is not None else None,
                input_boxes=[boxes] if boxes is not None else None,
                input_labels=[labels] if labels is not None else None,
                return_tensors="pt",
            ).to(self.model.device, dtype=self.model.dtype)

            if return_image_embedding:
                if image_embedding is None:  # Compute image embeddings if not provided
                    image_embedding = self.model.get_image_embeddings(inputs["pixel_values"])

            if image_embedding is not None:
                if image_embedding.ndim == 3:
                    image_embedding = image_embedding.unsqueeze(0)
                inputs.pop("pixel_values", None)
                inputs.update({"image_embeddings": image_embedding.to(self.model.device, dtype=self.model.dtype)})

            outputs = self.model(
                **inputs, num_multimask_outputs=num_multimask_outputs, multimask_output=multimask_output, **kwargs
            )

            masks = (
                self.processor.image_processor.post_process_masks(
                    outputs.pred_masks, inputs["original_sizes"], inputs["reshaped_input_sizes"]
                )
            )[0].cpu()
            return ImageMaskGenerationOutput(
                masks=[
                    [CompressedRLE(**encode_mask_to_rle(mask)) for mask in prediction_masks]
                    for prediction_masks in masks
                ],
                scores=NDArrayFloat.from_torch(outputs.iou_scores[0].cpu()),
                image_embedding=(
                    NDArrayFloat.from_torch(image_embedding[0].cpu()) if return_image_embedding else None
                ),
            )

    def text_image_conditional_generation(
        self,
        prompt: str | list[dict[str, Any]],
        images: list["Tensor"],
        generation_config: "GenerationConfig" | None = None,
        **kwargs: Any,
    ) -> TextImageConditionalGenerationOutput:
        """Generate text from an image and a prompt.

        Args:
            prompt: Prompt for the generation.
            images: Images for the generation.
            generation_config: Configuration for the generation as Hugging Face's GenerationConfig.
            kwargs: Additional keyword arguments.
        """
        with torch.inference_mode():
            if generation_config is None:
                generation_config = GenerationConfig()

            generation_config = self._fill_generation_config(generation_config, **kwargs)

            if isinstance(prompt, list):
                prompt = self.processor.apply_chat_template(prompt, add_generation_prompt=True)

            inputs = self.processor(prompt, images, return_tensors="pt").to(self.model.device)
            generate_ids = self.model.generate(**inputs, generation_config=generation_config)

            total_tokens: int = generate_ids.shape[1]
            prompt_tokens: int = inputs["input_ids"].shape[1]
            completion_tokens: int = total_tokens - prompt_tokens

            output = self.processor.decode(
                generate_ids[0, prompt_tokens:], skip_special_tokens=True, clean_up_tokenization_spaces=False
            )

            return TextImageConditionalGenerationOutput(
                generated_text=output,
                usage=UsageConditionalGeneration(
                    prompt_tokens=prompt_tokens,
                    completion_tokens=completion_tokens,
                    total_tokens=total_tokens,
                ),
                generation_config=generation_config.to_diff_dict(),
            )

    def image_zero_shot_detection(
        self,
        image: "Tensor" | Image,
        classes: str,
        box_threshold: float,
        text_threshold: float,
        **kwargs: Any,
    ) -> ImageZeroShotDetectionOutput:
        """Perform zero shot detection on an image.

        Args:
            image: The image.
            classes: The list of classes to detect in the format 'class1. class2'.
            box_threshold: The threshold for bounding boxes detection.
            text_threshold: The threshold for the classes identification during zero shot learning phase.
            kwargs: Additional arguments.

        Returns:
            The output of image zero-shot detection task.
        """
        with torch.inference_mode():
            inputs = self.processor(images=image, text=classes, return_tensors="pt").to(self.model.device)

            outputs = self.model(**inputs)

            target_size = (
                (image.shape[-2], image.shape[-1]) if isinstance(image, torch.Tensor) else (image.height, image.width)
            )

            result = self.processor.post_process_grounded_object_detection(
                outputs,
                inputs.input_ids,
                box_threshold=box_threshold,
                text_threshold=text_threshold,
                target_sizes=[target_size],
            )[0]

            return ImageZeroShotDetectionOutput(
                boxes=[[int(round(x, 0)) for x in box.tolist()] for box in result["boxes"]],
                scores=result["scores"],
                classes=result["labels"],
            )



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

import pytest

from pixano_inference.client import PixanoInferenceClient


@pytest.fixture(scope="session")
def simple_pixano_inference_client() -> PixanoInferenceClient:
    return PixanoInferenceClient(
        url="http://localhost:8081",
        app_name="Pixano Inference",
        app_version="0.1.0",
        app_description="Pixano Inference",
        num_cpus=4,
        num_gpus=2,
        num_nodes=1,
        gpus_used=[],
    )



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Tasks module."""

from enum import Enum


class Task(Enum):
    """Task base class."""

    pass



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""API routes for model providers."""

from typing import Annotated

from fastapi import APIRouter, Body, Depends, HTTPException

from pixano_inference.celery import add_celery_worker_and_queue, celery_app, delete_celery_worker_and_queue
from pixano_inference.providers.utils import instantiate_provider
from pixano_inference.pydantic import CeleryTask, ModelConfig
from pixano_inference.settings import Settings, get_pixano_inference_settings
from pixano_inference.tasks.utils import is_task


router = APIRouter(prefix="/providers", tags=["models"])


@router.post("/instantiate", response_model=CeleryTask)
async def instantiate_model(
    config: ModelConfig,
    provider: Annotated[str, Body()],
    settings: Annotated[Settings, Depends(get_pixano_inference_settings)],
):
    """Instantiate a model from a provider.

    Args:
        config: Model configuration for instantiation.
        provider: The model provider.
        settings: Settings of the app.
    """
    try:
        instantiate_provider(provider)
    except ValueError as e:
        raise HTTPException(status_code=404, detail=f"Provider {provider} does not exist.") from e
    except ImportError as e:
        raise HTTPException(status_code=500, detail=f"Provider {provider} is not installed.") from e

    if not is_task(config.task):
        raise HTTPException(status_code=404, detail=f"Task {config.task} is not a supported task.")

    gpu = settings.add_model(config.name, config.task)
    try:
        task = add_celery_worker_and_queue(model_config=config, provider=provider, gpu=gpu)
    except Exception as e:
        settings.remove_model(config.name)
        delete_celery_worker_and_queue(config.name)
        raise HTTPException(status_code=400, detail=f"Error while instantiating {config.task} - {str(e)}") from e
    return task


@router.get("/instantiate/{task_id}", response_model=CeleryTask)
async def get_instantiate_model_status(
    task_id: str,
) -> CeleryTask:
    """Return status of model instantiation."""
    task_result = celery_app.AsyncResult(task_id)
    return CeleryTask(id=task_result.id, status=task_result.status)


@router.delete("/model/{model_name}")
async def delete_model(
    model_name: str,
    settings: Annotated[Settings, Depends(get_pixano_inference_settings)],
):
    """Delete a model from the system.

    Args:
        model_name: The name of the model to be deleted.
        settings: Settings of the app.
    """
    if model_name not in settings.models:
        raise HTTPException(status_code=404, detail=f"Model {model_name} not found")
    settings.remove_model(model_name)
    delete_celery_worker_and_queue(model_name)
    return



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

from typing import TYPE_CHECKING, Any

import pytest

from pixano_inference.models import BaseInferenceModel
from pixano_inference.pydantic.tasks.multimodal.conditional_generation import (
    TextImageConditionalGenerationOutput,
)


if TYPE_CHECKING:
    from torch import Tensor


class MockModel(BaseInferenceModel):
    def __init__(self, name: str, provider: str):
        super().__init__(name, provider)

    @property
    def metadata(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "provider": self.provider,
        }

    def delete(self):
        pass

    def image_mask_generation(
        self, image: "Tensor", points: list[list[int]], boxes: list[list[int]], **kwargs: Any
    ) -> "Tensor":
        return ...

    def text_image_conditional_generation(
        self, prompt: str, image: "Tensor", **kwargs: Any
    ) -> TextImageConditionalGenerationOutput:
        pass


@pytest.fixture(scope="module")
def model():
    return MockModel("test_model", "test_provider")


class TestBaseInferenceModel:
    def test_metadata(self, model: MockModel):
        assert model.metadata == {
            "name": "test_model",
            "provider": "test_provider",
        }



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

# ================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# ================================

"""Generate the code API reference pages

Refer to https://mkdocstrings.github.io/recipes/ for more details.
"""

from pathlib import Path

import mkdocs_gen_files


SRC_PATH = "./src/pixano_inference"
REF_PATH = "api_reference"
IGNORED_FILES = ["__init__.py", "__version__.py"]

nav = mkdocs_gen_files.Nav()
for path in sorted(Path(SRC_PATH).rglob("*.py")):
    module_path = path.relative_to(SRC_PATH).with_suffix("")
    doc_path = path.relative_to(SRC_PATH).with_suffix(".md")
    full_doc_path = Path(REF_PATH, doc_path)

    parts = list(module_path.parts)

    if parts[-1] == "__init__":
        parts = parts[:-1]
    elif parts[-1] == "__main__":
        continue

    # Create root index file
    if not parts:
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
        nav["index"] = doc_path.as_posix()
        with open(f"./docs/{REF_PATH}/index.md", "r", encoding="utf-8") as index_file:
            print(index_file)
            lines = index_file.readlines()
            with mkdocs_gen_files.open(full_doc_path, "w") as fd:
                fd.writelines(lines)
        mkdocs_gen_files.set_edit_path(full_doc_path, path)

    # Add all python files that are not in the ignore list
    elif not any(ignored in path.name for ignored in IGNORED_FILES):
        nav[parts] = doc_path.as_posix()
        with mkdocs_gen_files.open(full_doc_path, "w") as fd:
            print("::: pixano_inference." + ".".join(parts), file=fd)
        mkdocs_gen_files.set_edit_path(full_doc_path, path)


with mkdocs_gen_files.open(f"{REF_PATH}/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Provider for vLLM models."""

from typing import Any

from pixano_inference.models.vllm import VLLMModel
from pixano_inference.pydantic.tasks.multimodal.conditional_generation import (
    TextImageConditionalGenerationOutput,
    TextImageConditionalGenerationRequest,
)
from pixano_inference.tasks import Task, str_to_task
from pixano_inference.utils.package import (
    assert_vllm_installed,
    is_torch_installed,
)

from .base import ModelProvider
from .registry import register_provider


if is_torch_installed():
    import torch


@register_provider("vllm")
class VLLMProvider(ModelProvider):
    """Provider for vLLM models."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the vLLM provider."""
        assert_vllm_installed()
        super().__init__(*args, **kwargs)

    def load_model(
        self,
        name: str,
        task: Task | str,
        device: "torch.device",
        path: str | None = None,  # type: ignore[override]
        processor_config: dict = {},
        config: dict = {},
    ) -> VLLMModel:
        """Load a model from vLLM.

        Args:
            name: Name of the model.
            task: Task of the model.
            device: Device to use for the model.
            path: Path to the model or its Hugging Face hub's identifier.
            processor_config: Configuration for the processor.
            config: Configuration for the model.

        Returns:
            Loaded model.
        """
        if path is None:
            raise ValueError("Path is required to load a model from vLLm.")
        if isinstance(task, str):
            task = str_to_task(task)

        our_model = VLLMModel(
            name=name, vllm_model=path, model_config=config, processor_config=processor_config, device=device
        )

        return our_model

    def text_image_conditional_generation(
        self,
        request: TextImageConditionalGenerationRequest,
        model: VLLMModel,  # type: ignore[override]
        *args: Any,
        **kwargs: Any,
    ) -> TextImageConditionalGenerationOutput:
        """Generate text from an image and a prompt.

        Args:
            request: Request for text-image conditional generation.
            model: Model for text-image conditional generation
            args: Additional arguments.
            kwargs: Additional keyword arguments.

        Returns:
            Output of text-image conditional generation.
        """
        model_input = request.to_input()
        if model_input.images is not None:
            raise ValueError("images should be passed in the prompt for vLLM.")
        if isinstance(model_input.prompt, str):
            raise ValueError("Pixano-inference only support a chat template for vLLM.")

        output = model.text_image_conditional_generation(**model_input.model_dump(exclude="images"))
        return output



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""API routes for NLP tasks."""

from typing import Annotated

from fastapi import APIRouter, Depends

from pixano_inference.pydantic import (
    CeleryTask,
    TextImageConditionalGenerationRequest,
    TextImageConditionalGenerationResponse,
)
from pixano_inference.routers.utils import delete_task, execute_task_request, get_task_result
from pixano_inference.settings import Settings, get_pixano_inference_settings
from pixano_inference.tasks import MultimodalImageNLPTask


router = APIRouter(prefix="/multimodal", tags=["Multi-Modal"])


@router.post("/text-image/conditional_generation/", response_model=CeleryTask)
async def text_image_conditional_generation(
    request: TextImageConditionalGenerationRequest,
    settings: Annotated[Settings, Depends(get_pixano_inference_settings)],
) -> CeleryTask:
    """Generate text from an image and a prompt.

    Args:
        request: Request for text-image conditional generation.
        settings: Settings of the app.

    Returns:
        Response for text-image conditional generation.
    """
    return await execute_task_request(
        request=request, task=MultimodalImageNLPTask.CONDITIONAL_GENERATION, settings=settings
    )


@router.get(
    "/text-image/conditional_generation/{task_id}", response_model=TextImageConditionalGenerationResponse | CeleryTask
)
async def get_text_image_conditional_generation(task_id: str) -> TextImageConditionalGenerationResponse | CeleryTask:
    """Get the result of a text image conditional generation task.

    Args:
        task_id: ID of the task to retrieve.

    Returns:
        Response for text image conditional generation.
    """
    result = await get_task_result(task_id, response_type=TextImageConditionalGenerationResponse)
    return result


@router.delete("/text-image/conditional_generation/{task_id}", response_model=None)
async def delete_text_image_conditional_generation(task_id: str) -> None:
    """Delete a text image conditional generation task."""
    return await delete_task(task_id=task_id)



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

# ruff: noqa: F401
# ruff: noqa: D104

from .conditional_generation import (
    TextImageConditionalGenerationInput,
    TextImageConditionalGenerationOutput,
    TextImageConditionalGenerationRequest,
    TextImageConditionalGenerationResponse,
    UsageConditionalGeneration,
)



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Pydantic model for the Vector databases."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import numpy as np
from pydantic import BaseModel, model_validator
from typing_extensions import Self

from pixano_inference.data.read_vector_databases import read_lance_vector


if TYPE_CHECKING:
    from torch import Tensor


class LanceVector(BaseModel):
    """Lance vector model.

    Attributes:
        path: The path to the LANCE dataset.
        column: The column to read.
        indice: The index of the row to read.
        where: The filter to apply. If specified, only one row should be returned.
        shape: The shape of the vector.
    """

    path: str
    column: str
    indice: int | None = None
    where: str | None = None
    shape: list[int] | None = None

    @model_validator(mode="after")
    def validate_indice_or_where(self) -> Self:
        """Validates that only one of 'indice' and 'where' is specified."""
        if self.indice is not None and self.where is not None:
            raise ValueError("Only one of 'indice' and 'where' can be specified.")
        elif self.indice is None and self.where is None:
            raise ValueError("One of 'indice' and 'where' must be specified.")
        return self

    def read_vector(self, return_type: Literal["tensor", "numpy"] = "tensor") -> "Tensor" | np.ndarray:
        """Reads a vector from a Lance dataset.

        Args:
            return_type: The type of the return value. Either 'tensor' or 'numpy'.

        Returns:
            The vector.
        """
        return read_lance_vector(self.path, self.column, self.indice, self.where, self.shape, return_type)



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Provider for the SAM2 model."""

from .base import BaseProvider
from .registry import PROVIDERS_REGISTRY


def get_provider(provider: str) -> type[BaseProvider]:
    """Return the provider from the registry."""
    if (actual_provider := PROVIDERS_REGISTRY.get(provider)) is not None:
        return actual_provider
    raise ValueError(f"Provider {provider} not found.")


def get_providers() -> list[str]:
    """Return the list of providers in the registry."""
    return list(PROVIDERS_REGISTRY.keys())


def is_provider(provider: str) -> bool:
    """Return True if the provider is in the registry."""
    return provider in PROVIDERS_REGISTRY


def get_provider_name(provider: BaseProvider) -> str:
    """Return the name of the provider."""
    for name, provider_cls in PROVIDERS_REGISTRY.items():
        if isinstance(provider, provider_cls):
            return name
    raise ValueError(f"Provider {provider} not found.")


def instantiate_provider(provider: str) -> BaseProvider:
    """Instantiate a provider."""
    return get_provider(provider)()



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Settings for the Pixano Inference API."""

import os
from typing import Any

from dotenv import load_dotenv
from pydantic_settings import BaseSettings

from pixano_inference.__version__ import __version__
from pixano_inference.utils.package import is_torch_installed


load_dotenv()


if is_torch_installed():
    import torch


class Settings(BaseSettings):
    """Application settings.

    Attributes:
        app_name: The name of the application.
        app_version: The version of the application.
        app_description: A description of the application.
        num_cpus: The number of CPUs accessible to the application.
        num_gpus: The number of GPUs available for inference.
        num_nodes: The number of nodes available for inference.
        gpus_used: The list of GPUs used by the application.
    """

    app_name: str = "Pixano Inference"
    app_version: str = __version__
    app_description: str = "Pixano Inference API for multimodal tasks."
    num_cpus: int
    num_gpus: int
    num_nodes: int = 1  # TODO: for now only single node is supported
    gpus_used: list[int] = []
    gpu_to_model: dict[int, str] = {}
    models: list[str] = []
    models_to_task: dict[str, str] = {}

    def __init__(self, **data: Any):
        """Initialize the settings."""
        if "num_cpus" not in data:
            data["num_cpus"] = os.cpu_count()
        if "num_gpus" not in data:
            if is_torch_installed():
                if torch.cuda.is_available():
                    data["num_gpus"] = torch.cuda.device_count()
                else:
                    data["num_gpus"] = 0
            else:
                data["num_gpus"] = 0

        super().__init__(**data)

    @property
    def gpus_available(self) -> list[int]:
        """Return the available GPUs."""
        return [i for i in range(self.num_gpus) if i not in self.gpus_used]

    def reserve_gpu(self) -> int | None:
        """Reserve a gpu if any available."""
        gpus = self.gpus_available
        if len(gpus) == 0:
            return None
        selected_gpu = gpus[0]
        self.gpus_used.append(selected_gpu)
        return selected_gpu

    def free_gpu(self, gpu: int) -> None:
        """Free a GPU if used."""
        try:
            self.gpus_used.remove(gpu)
        except ValueError:
            pass
        self.gpu_to_model.pop(gpu)
        return

    def add_model(self, model: str, task: str) -> int | None:
        """Add a model."""
        if model in self.models:
            raise ValueError(f"Model {model} already registered.")
        gpu = self.assign_model_gpu(model)
        self.models_to_task[model] = task
        self.models.append(model)
        return gpu

    def assign_model_gpu(self, model: str) -> int | None:
        """Assign a model to a gpu.

        Args:
            model: The model name.

        Returns:
            The gpu index. If no gpu available, returns None.
        """
        gpu = self.reserve_gpu()
        if gpu is not None:
            self.gpu_to_model[gpu] = model
        return gpu

    def remove_model(self, model: str) -> None:
        """Remove a model."""
        self.models.remove(model)
        self.models_to_task.pop(model, None)
        for gpu, model_stored in self.gpu_to_model.items():
            if model_stored == model:
                self.free_gpu(gpu)
                break
        return


def get_pixano_inference_settings() -> Settings:
    """Return the settings."""
    return PIXANO_INFERENCE_SETTINGS


PIXANO_INFERENCE_SETTINGS = Settings()



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Base classes for providers."""

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any

from pixano_inference.models.base import BaseInferenceModel
from pixano_inference.pydantic import (
    APIRequest,
    BaseResponse,
    ImageMaskGenerationOutput,
    ImageMaskGenerationRequest,
    ImageZeroShotDetectionOutput,
    ImageZeroShotDetectionRequest,
    TextImageConditionalGenerationOutput,
    TextImageConditionalGenerationRequest,
    VideoMaskGenerationOutput,
    VideoMaskGenerationRequest,
)
from pixano_inference.tasks import Task
from pixano_inference.utils.package import is_torch_installed


if is_torch_installed():
    import torch

if TYPE_CHECKING:
    from pixano_inference.models import BaseInferenceModel


class BaseProvider(ABC):
    """Base class for providers."""

    def __init__(self, **kwargs: Any):
        """Initialize the provider."""
        pass

    def image_mask_generation(
        self, request: ImageMaskGenerationRequest, model: BaseInferenceModel, *args: Any, **kwargs: Any
    ) -> ImageMaskGenerationOutput:
        """Generate a mask from the image.

        Args:
            request: Request for the generation.
            model: Model to use for the generation.
            args: Additional arguments.
            kwargs: Additional keyword arguments.
        """
        raise NotImplementedError("This provider does not support image mask generation.")

    def text_image_conditional_generation(
        self, request: TextImageConditionalGenerationRequest, model: BaseInferenceModel, *args: Any, **kwargs: Any
    ) -> TextImageConditionalGenerationOutput:
        """Generate an image from the text and image.

        Args:
            request: Request for the generation.
            model: Model to use for the generation.
            args: Additional arguments.
            kwargs: Additional keyword arguments.

        Returns:
            Output of the generation
        """
        raise NotImplementedError("This provider does not support text-image conditional generation.")

    def video_mask_generation(
        self, request: VideoMaskGenerationRequest, model: BaseInferenceModel, *args: Any, **kwargs: Any
    ) -> VideoMaskGenerationOutput:
        """Generate a mask from the video.

        Args:
            request: Request for the generation.
            model: Model to use for the generation.
            args: Additional arguments.
            kwargs: Additional keyword arguments.

        Returns:
            Output of the generation
        """
        raise NotImplementedError("This provider does not support video mask generation.")

    def image_zero_shot_detection(
        self, request: ImageZeroShotDetectionRequest, model: BaseInferenceModel, *args: Any, **kwargs: Any
    ) -> ImageZeroShotDetectionOutput:
        """Perform zero-shot image detection."""
        raise NotImplementedError("This provider does not support image zero-shot detection.")


class APIProvider(BaseProvider):
    """Base class for API providers."""

    @property
    @abstractmethod
    def api_url(self) -> str:
        """URL of the API."""
        ...

    @abstractmethod
    def send_request(self, request: APIRequest) -> BaseResponse:
        """Send a request to the API and return the response.

        Args:
            request: Request to send.

        Returns:
            Response from the API.
        """
        ...


class ModelProvider(BaseProvider):
    """Base class for model providers."""

    @abstractmethod
    def load_model(
        self,
        name: str,
        task: Task | str,
        device: "torch.device",
        path: Path | str | None = None,
        processor_config: dict = {},
        config: dict = {},
    ) -> "BaseInferenceModel":
        """Load the model from the provider.

        Args:
            name: Name of the model.
            task: Task of the model.
            device: Device to use for the model.
            path: Path to the model.
            processor_config: Processor configuration.
            config: Configuration for the model.

        Returns:
            The loaded model.
        """
        ...



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Inference models for vLLM."""

from __future__ import annotations

import gc
from typing import Any

from pixano_inference.models.base import BaseInferenceModel
from pixano_inference.pydantic.tasks.multimodal.conditional_generation import (
    TextImageConditionalGenerationOutput,
    UsageConditionalGeneration,
)
from pixano_inference.utils.package import (
    assert_vllm_installed,
    is_torch_installed,
    is_vllm_installed,
)


if is_torch_installed():
    import torch

if is_vllm_installed():
    import msgspec
    import vllm
    from vllm import LLM, CompletionOutput, RequestOutput, SamplingParams


class VLLMModel(BaseInferenceModel):
    """Inference model for vLLM."""

    def __init__(
        self,
        name: str,
        vllm_model: str,
        model_config: dict[str, Any],
        processor_config: dict[str, Any],
        device: torch.device | str | None = None,
    ):
        """Initialize the model.

        Args:
            name: Name of the model.
            vllm_model: The model Hugging Face hub's identifier.
            model_config: Configuration for the model.
            processor_config: Configuration for the processor of the model.
            device: The device to use for inference.
        """
        assert_vllm_installed()

        super().__init__(name, provider="vllm")
        self.vllm_model = vllm_model
        self.model_config = model_config
        self.processor_config = processor_config
        self.model = LLM(model=vllm_model, **model_config, **processor_config, device=device)

    @property
    def metadata(self) -> dict[str, Any]:
        """Return the metadata of the model."""
        return {
            "name": self.name,
            "vllm_model": self.vllm_model,
            "provider": self.provider,
            "provider_version": vllm.__version__,
            "processor_config": self.processor_config,
            "model_config": self.model_config,
        }

    def delete(self):
        """Delete the model."""
        del self.model
        gc.collect()
        torch.cuda.empty_cache()

    def text_image_conditional_generation(
        self,
        prompt: list[dict[str, Any]],
        temperature: float = 1.0,
        max_new_tokens: int = 16,
        **kwargs: Any,
    ) -> TextImageConditionalGenerationOutput:
        """Generate text from an image and a prompt from the vLLM's `LLM.chat` method.

        Args:
            prompt: Prompt for the generation.
            temperature: Temperature for the generation.
            max_new_tokens: Maximum number of tokens to generate.
            kwargs: Additional generation arguments.
        """
        sampling_params = SamplingParams(temperature=temperature, max_tokens=max_new_tokens, **kwargs)
        with torch.inference_mode():
            request_output: RequestOutput = self.model.chat(
                messages=prompt, use_tqdm=False, sampling_params=sampling_params
            )[0]
            prompt = request_output.prompt
            output: CompletionOutput = request_output.outputs[0]

            prompt_tokens = len(request_output.prompt_token_ids)
            completion_tokens = len(output.token_ids)
            total_tokens = prompt_tokens + completion_tokens

            return TextImageConditionalGenerationOutput(
                generated_text=output.text,
                usage=UsageConditionalGeneration(
                    prompt_tokens=prompt_tokens,
                    completion_tokens=completion_tokens,
                    total_tokens=total_tokens,
                ),
                generation_config=msgspec.to_builtins(sampling_params),
            )



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

# ruff: noqa: F401
# ruff: noqa: D104

from .__version__ import __version__
from .settings import PIXANO_INFERENCE_SETTINGS, get_pixano_inference_settings



# =================================
# Copyright: CEA-LIST/DIASI/SIALV
# Author : pixano@cea.fr
# License: CECILL-C
# =================================

"""Pydantic models for image tasks."""

from typing import Any

import numpy as np
from PIL.Image import Image
from pycocotools import mask as mask_api
from pydantic import BaseModel, field_serializer, field_validator, model_validator


def mask_to_rle(mask: Image | np.ndarray) -> dict:
    """Encode mask from Pillow or NumPy array to RLE.

    Args:
        mask: Mask as Pillow or NumPy array.

    Returns:
        Mask as RLE.
    """
    mask_array = np.asfortranarray(mask)
    return mask_api.encode(mask_array)


def rle_to_mask(rle: dict[str, list[int] | bytes]) -> np.ndarray:
    """Decode mask from RLE to NumPy array.

    Args:
        rle: Mask as RLE.

    Returns:
        Mask as NumPy array.
    """
    return mask_api.decode(rle)


class CompressedRLE(BaseModel):
    """Compressed RLE mask type.

    Attributes:
        size: Mask size.
        counts: Mask RLE encoding.
    """

    size: list[int]
    counts: bytes

    @field_validator("counts", mode="before")
    @classmethod
    def _validate_counts(cls, value: bytes | str) -> bytes:
        if isinstance(value, str):
            value = bytes(value, "utf-8")
        return value

    @model_validator(mode="after")
    def _validate_fields(self):
        if (
            len(self.size) != 2
            and not all(isinstance(s, int) and s > 0 for s in self.size)
            and not (self.size == [0, 0] and self.counts == b"")
        ):
            raise ValueError("Mask size must have 2 elements and be positive integers or [0, 0] for empty mask.")
        return self

    @field_serializer("counts")
    def _serialize_counts(self, value: bytes) -> str:
        return str(value, "utf-8")

    @staticmethod
    def from_mask(mask: Image | np.ndarray, **kwargs: Any) -> "CompressedRLE":
        """Create a compressed RLE mask from a NumPy array.

        Args:
            mask: The mask as a NumPy array.
            kwargs: Additional arguments.

        Returns:
            The compressed RLE mask.
        """
        rle = mask_to_rle(mask)
        return CompressedRLE(size=rle["size"], counts=rle["counts"], **kwargs)

    def to_mask(self) -> np.ndarray:
        """Convert the compressed RLE mask to a NumPy array.

        Returns:
            The mask as a NumPy array.
        """
        return rle_to_mask({"size": self.size, "counts": self.counts})
